import os
import numpy as np
import pickle

from autofit import exc
from autofit.non_linear.optimize.abstract_optimize import AbstractOptimizer
from autofit.non_linear.samples import OptimizerSamples

from autofit.non_linear.log import logger


class AbstractPySwarms(AbstractOptimizer):
    def __init__(
        self,
        paths=None,
        prior_passer=None,
        n_particles=None,
        iters=None,
        cognitive=None,
        social=None,
        inertia=None,
        ftol=None,
        initializer=None,
        iterations_per_update=None,
        number_of_cores=None,
    ):
        """
        A PySwarms Particle Swarm Optimizer global non-linear search.

        For a full description of PySwarms, checkout its Github and readthedocs webpages:

        https://github.com/ljvmiranda921/pyswarms
        https://pyswarms.readthedocs.io/en/latest/index.html

        A Global-best Particle Swarm Optimization (gbest PSO) algorithm.

        It takes a set of candidate solutions, and tries to find the best solution using a position-velocity update
        method. Uses a star-topology where each particle is attracted to the best performing particle.

        The position update can be defined as:
        xi(t+1)=xi(t)+vi(t+1)

        Where the position at the current timestep t is updated using the computed velocity at t+1.

        Furthermore, the velocity update is defined as:

        vij(t+1)=w∗vij(t)+c1r1j(t)[yij(t)−xij(t)]+c2r2j(t)[y^j(t)−xij(t)]

        Here, c1 and c2 are the cognitive and social parameters respectively. They control the particle’s behavior
        given two choices: (1) to follow its personal best or (2) follow the swarm’s global best position. Overall,
        this dictates if the swarm is explorative or exploitative in nature. In addition, a parameter w controls
        the inertia of the swarm’s movement.

        Extensions:

        Allows runs to be terminated and resumed from the point it was terminated. This is achieved by outputting
        the necessary results (e.g. the points of the particles) during the model-fit after an input number of
        iterations.

        Different options for particle intialization, with the default 'prior' method starting all particles over
        the priors defined by each parameter.

        If you use *PySwarms* as part of a published work, please cite the package following the instructions under the
        *Attribution* section of the GitHub page.

        All remaining attributes are emcee parameters and described at the PySwarms API webpage:

        https://pyswarms.readthedocs.io/en/latest/index.html

        Parameters
        ----------
        paths : af.Paths
            Manages all paths, e.g. where the search outputs are stored, the samples, backups, etc.
        prior_passer : af.PriorPasser
            Controls how priors are passed from the results of this non-linear search to a subsequent non-linear search.
        n_particles : int
            The number of particles in the swarm used to sample parameter space.
        iters : int
            The number of iterations that are used to sample parameter space.
        cognitive : float
            The cognitive parameter defining how the PSO particles interact with one another to sample parameter space.
        social : float
            The social parameter defining how the PSO particles interact with one another to sample parameter space.
        inertia : float
            The inertia parameter defining how the PSO particles interact with one another to sample parameter space.
        ftol : float
            Relative error in objective_func(best_pos) acceptable for convergence.
        initializer : non_linear.initializer.Initializer
            Generates the initialize samples of non-linear parameter space (see autofit.non_linear.initializer).
        number_of_cores : int
            The number of cores Emcee sampling is performed using a Python multiprocessing Pool instance. If 1, a
            pool instance is not created and the job runs in serial.
        """

        self.n_particles = (
            self._config("search", "n_particles", int)
            if n_particles is None
            else n_particles
        )
        self.iters = self._config("search", "iters", int) if iters is None else iters

        self.cognitive = (
            self._config("search", "cognitive", float)
            if cognitive is None
            else cognitive
        )
        self.social = (
            self._config("search", "social", float) if social is None else social
        )
        self.inertia = (
            self._config("search", "inertia", float) if inertia is None else inertia
        )
        self.ftol = self._config("search", "ftol", float) if ftol is None else ftol

        super().__init__(
            paths=paths,
            prior_passer=prior_passer,
            initializer=initializer,
            iterations_per_update=iterations_per_update,
        )

        self.number_of_cores = (
            self._config("parallel", "number_of_cores", int)
            if number_of_cores is None
            else number_of_cores
        )

        logger.debug("Creating PySwarms NLO")

    class Fitness(AbstractOptimizer.Fitness):
        def __call__(self, parameters):

            figures_of_merit = []

            for params_of_particle in parameters:

                try:
                    figure_of_merit = self.figure_of_merit_from_parameters(
                        parameters=params_of_particle
                    )
                except exc.FitException:
                    figure_of_merit = -2.0 * self.resample_figure_of_merit

                figures_of_merit.append(figure_of_merit)

            return np.asarray(figures_of_merit)

        def figure_of_merit_from_parameters(self, parameters):
            """The figure of merit is the value that the non-linear search uses to sample parameter space. *PySwarms*
            uses the chi-squared value, which is the -2.0*log_posterior."""
            try:
                return -2.0 * self.log_posterior_from_parameters(parameters=parameters)
            except exc.FitException:
                raise exc.FitException

    def _fit(self, model, analysis):
        """
        Fit a model using PySwarms and the Analysis class which contains the data and returns the log likelihood from
        instances of the model, which the non-linear search seeks to maximize.

        Parameters
        ----------
        model : ModelMapper
            The model which generates instances for different points in parameter space.
        analysis : Analysis
            Contains the data and the log likelihood function which fits an instance of the model to the data, returning
            the log likelihood the non-linear search maximizes.

        Returns
        -------
        A result object comprising the Samples object that inclues the maximum log likelihood instance and full
        chains used by the fit.
        """
        pool, pool_ids = self.make_pool()

        fitness_function = self.fitness_function_from_model_and_analysis(
            model=model, analysis=analysis, pool_ids=pool_ids
        )

        if os.path.exists("{}/{}.pickle".format(self.paths.samples_path, "points")):

            init_pos = self.load_points[-1]
            total_iterations = self.load_total_iterations

            logger.info("Existing PySwarms samples found, resuming non-linear search.")

        else:

            initial_unit_parameters, initial_parameters, initial_log_posteriors = self.initializer.initial_samples_from_model(
                total_points=self.n_particles,
                model=model,
                fitness_function=fitness_function,
            )

            init_pos = np.zeros(shape=(self.n_particles, model.prior_count))

            for index, parameters in enumerate(initial_parameters):

                init_pos[index, :] = np.asarray(parameters)

            total_iterations = 0

            logger.info("No PySwarms samples found, beginning new non-linear search. ")

        lower_bounds = []
        upper_bounds = []

        for key, value in model.prior_class_dict.items():
            lower_bounds.append(key.lower_limit)
            upper_bounds.append(key.upper_limit)

        bounds = (np.asarray(lower_bounds), np.asarray(upper_bounds))

        logger.info("Running PySwarmsGlobal Optimizer...")

        while total_iterations < self.iters:

            pso = self.sampler_fom_model_and_fitness(
                model=model,
                fitness_function=fitness_function,
                bounds=bounds,
                init_pos=init_pos,
            )

            iterations_remaining = self.iters - total_iterations

            if self.iterations_per_update > iterations_remaining:
                iterations = iterations_remaining
            else:
                iterations = self.iterations_per_update

            if iterations > 0:

                pso.optimize(objective_func=fitness_function.__call__, iters=iterations)

                total_iterations += iterations

                with open(
                    f"{self.paths.samples_path}/total_iterations.pickle", "wb"
                ) as f:
                    pickle.dump(total_iterations, f)

                with open(f"{self.paths.samples_path}/points.pickle", "wb") as f:
                    pickle.dump(pso.pos_history, f)

                with open(
                    f"{self.paths.samples_path}/log_posteriors.pickle", "wb"
                ) as f:
                    pickle.dump([-0.5 * cost for cost in pso.cost_history], f)

                self.perform_update(
                    model=model, analysis=analysis, during_analysis=True
                )

                init_pos = self.load_points[-1]

        logger.info("PySwarmsGlobal complete")

    @property
    def tag(self):
        """Tag the output folder of the PySwarms non-linear search, according to the number of particles and
        parameters defining the search strategy."""

        name_tag = self._config("tag", "name", str)
        n_particles_tag = f"{self._config('tag', 'n_particles')}_{self.n_particles}"
        cognitive_tag = f"{self._config('tag', 'cognitive')}_{self.cognitive}"
        social_tag = f"{self._config('tag', 'social')}_{self.social}"
        inertia_tag = f"{self._config('tag', 'inertia')}_{self.inertia}"

        return f"{name_tag}[{n_particles_tag}_{cognitive_tag}_{social_tag}_{inertia_tag}]"

    def copy_with_name_extension(self, extension, remove_phase_tag=False):
        """Copy this instance of the emcee non-linear search with all associated attributes.

        This is used to set up the non-linear search on phase extensions."""
        copy = super().copy_with_name_extension(
            extension=extension, remove_phase_tag=remove_phase_tag
        )
        copy.prior_passer = self.prior_passer
        copy.n_particles = self.n_particles
        copy.iters = self.iters
        copy.cognitive = self.cognitive
        copy.social = self.social
        copy.intertia = self.inertia
        copy.ftol = self.ftol
        copy.initializer = self.initializer
        copy.iterations_per_update = self.iterations_per_update
        copy.number_of_cores = self.number_of_cores

        return copy

    def fitness_function_from_model_and_analysis(self, model, analysis, pool_ids=None):

        return PySwarmsGlobal.Fitness(
            paths=self.paths,
            model=model,
            analysis=analysis,
            samples_from_model=self.samples_via_sampler_from_model,
            pool_ids=pool_ids,
        )

    def sampler_fom_model_and_fitness(self, model, fitness_function):
        raise NotImplementedError()

    def samples_via_sampler_from_model(self, model):
        """Create an *OptimizerSamples* object from this non-linear search's output files on the hard-disk and model.

        For PySwarms, all quantities are extracted via pickled states of the particle and cost histories.

        Parameters
        ----------
        model
            The model which generates instances for different points in parameter space. This maps the points from unit
            cube values to physical values via the priors.
        """

        parameters = [
            param.tolist() for parameters in self.load_points for param in parameters
        ]
        log_priors = [
            sum(model.log_priors_from_vector(vector=vector)) for vector in parameters
        ]
        log_posteriors = self.load_log_posteriors
        log_likelihoods = [lp - prior for lp, prior in zip(log_posteriors, log_priors)]
        weights = len(log_likelihoods)*[1.0]

        return OptimizerSamples(
            model=model,
            parameters=[parameters.tolist()[0] for parameters in self.load_points],
            log_likelihoods=log_likelihoods,
            log_priors=log_priors,
            weights=weights,
            time=self.timer.time
        )

    @property
    def load_total_iterations(self):
        with open(
            "{}/{}.pickle".format(self.paths.samples_path, "total_iterations"), "rb"
        ) as f:
            return pickle.load(f)

    @property
    def load_points(self):
        with open("{}/{}.pickle".format(self.paths.samples_path, "points"), "rb") as f:
            return pickle.load(f)

    @property
    def load_log_posteriors(self):
        with open(
            "{}/{}.pickle".format(self.paths.samples_path, "log_posteriors"), "rb"
        ) as f:
            return pickle.load(f)


class PySwarmsGlobal(AbstractPySwarms):
    def __init__(
        self,
        paths=None,
        prior_passer=None,
        n_particles=None,
        iters=None,
        cognitive=None,
        social=None,
        inertia=None,
        ftol=None,
        initializer=None,
        iterations_per_update=None,
        number_of_cores=None,
    ):
        """ A PySwarms Particle Swarm Optimizer global non-linear search.

        For a full description of PySwarms, checkout its Github and readthedocs webpages:

        https://github.com/ljvmiranda921/pyswarms

        https://pyswarms.readthedocs.io/en/latest/index.html

        A Global-best Particle Swarm Optimization (gbest PSO) algorithm.

        It takes a set of candidate solutions, and tries to find the best solution using a position-velocity update
        method. Uses a star-topology where each particle is attracted to the best performing particle.

        The position update can be defined as:
        xi(t+1)=xi(t)+vi(t+1)

        Where the position at the current timestep t is updated using the computed velocity at t+1.

        Furthermore, the velocity update is defined as:

        vij(t+1)=w∗vij(t)+c1r1j(t)[yij(t)−xij(t)]+c2r2j(t)[y^j(t)−xij(t)]

        Here, c1 and c2 are the cognitive and social parameters respectively. They control the particle’s behavior
        given two choices: (1) to follow its personal best or (2) follow the swarm’s global best position. Overall,
        this dictates if the swarm is explorative or exploitative in nature. In addition, a parameter w controls
        the inertia of the swarm’s movement.

        Extensions:

        - Allows runs to be terminated and resumed from the point it was terminated. This is achieved by outputting the
          necessary results (e.g. the points of the particles) during the model-fit after an input number of iterations.

        - Different options for particle intialization, with the default 'prior' method starting all particles over the
        priors defined by each parameter.

        If you use *PySwarms* as part of a published work, please cite the package following the instructions under the
        *Attribution* section of the GitHub page.

        Parameters
        ----------
        paths : af.Paths
            Manages all paths, e.g. where the search outputs are stored, the samples, backups, etc.
        prior_passer : af.PriorPasser
            Controls how priors are passed from the results of this non-linear search to a subsequent non-linear search.
        n_particles : int
            The number of particles in the swarm used to sample parameter space.
        iters : int
            The number of iterations that are used to sample parameter space.
        cognitive : float
            The cognitive parameter defining how the PSO particles interact with one another to sample parameter space.
        social : float
            The social parameter defining how the PSO particles interact with one another to sample parameter space.
        inertia : float
            The inertia parameter defining how the PSO particles interact with one another to sample parameter space.
        ftol : float
            Relative error in objective_func(best_pos) acceptable for convergence.
        initializer : non_linear.initializer.Initializer
            Generates the initialize samples of non-linear parameter space (see autofit.non_linear.initializer).
        number_of_cores : int
            The number of cores Emcee sampling is performed using a Python multiprocessing Pool instance. If 1, a
            pool instance is not created and the job runs in serial.

        All remaining attributes are emcee parameters and described at the PySwarms API webpage:

        https://pyswarms.readthedocs.io/en/latest/index.html
        """

        super().__init__(
            paths=paths,
            prior_passer=prior_passer,
            n_particles=n_particles,
            iters=iters,
            cognitive=cognitive,
            social=social,
            inertia=inertia,
            ftol=ftol,
            initializer=initializer,
            iterations_per_update=iterations_per_update,
            number_of_cores=number_of_cores,
        )

        logger.debug("Creating PySwarms NLO")

    def sampler_fom_model_and_fitness(self, model, fitness_function, bounds, init_pos):
        """Get the static Dynesty sampler which performs the non-linear search, passing it all associated input Dynesty
        variables."""

        import pyswarms

        return pyswarms.global_best.GlobalBestPSO(
            n_particles=self.n_particles,
            dimensions=model.prior_count,
            bounds=bounds,
            options={"c1": self.cognitive, "c2": self.social, "w": self.inertia},
            ftol=self.ftol,
            init_pos=init_pos,
        )


class PySwarmsLocal(AbstractPySwarms):
    def __init__(
        self,
        paths=None,
        prior_passer=None,
        n_particles=None,
        iters=None,
        cognitive=None,
        social=None,
        inertia=None,
        number_of_k_neighbors=None,
        minkowski_p_norm=None,
        ftol=None,
        initializer=None,
        iterations_per_update=None,
        number_of_cores=None,
    ):
        """ A PySwarms Particle Swarm Optimizer global non-linear search.

        For a full description of PySwarms, checkout its Github and readthedocs webpages:

        https://github.com/ljvmiranda921/pyswarms

        https://pyswarms.readthedocs.io/en/latest/index.html

        A Local-best Particle Swarm Optimization (lbest PSO) algorithm.

        Similar to global-best PSO, it takes a set of candidate solutions, and finds the best solution using a
        position-velocity update method. However, it uses a ring topology, thus making the particles attracted to its
         corresponding neighborhood.

        The position update can be defined as:
        xi(t+1)=xi(t)+vi(t+1)

        Where the position at the current timestep t
        is updated using the computed velocity at t+1

        . Furthermore, the velocity update is defined as:
        vij(t+1)=m∗vij(t)+c1r1j(t)[yij(t)−xij(t)]+c2r2j(t)[y^j(t)−xij(t)]

        However, in local-best PSO, a particle doesn’t compare itself to the overall performance of the swarm.
        Instead, it looks at the performance of its nearest-neighbours, and compares itself with them. In general,
        this kind of topology takes much more time to converge, but has a more powerful explorative feature.

        In this implementation, a neighbor is selected via a k-D tree imported from scipy. Distance are computed with
        either the L1 or L2 distance. The nearest-neighbours are then queried from this k-D tree. They are computed
        for every iteration.

        Extensions:

        - Allows runs to be terminated and resumed from the point it was terminated. This is achieved by outputting the
          necessary results (e.g. the points of the particles) during the model-fit after an input number of iterations.

        - Different options for particle intialization, with the default 'prior' method starting all particles over the
        priors defined by each parameter.

        If you use *PySwarms* as part of a published work, please cite the package following the instructions under the
        *Attribution* section of the GitHub page.

        Parameters
        ----------
        paths : af.Paths
            Manages all paths, e.g. where the search outputs are stored, the samples, backups, etc.
        prior_passer : af.PriorPasser
            Controls how priors are passed from the results of this non-linear search to a subsequent non-linear search.
        n_particles : int
            The number of particles in the swarm used to sample parameter space.
        iters : int
            The number of iterations that are used to sample parameter space.
        cognitive : float
            The cognitive parameter defining how the PSO particles interact with one another to sample parameter space.
        social : float
            The social parameter defining how the PSO particles interact with one another to sample parameter space.
        inertia : float
            The inertia parameter defining how the PSO particles interact with one another to sample parameter space.
        number_of_k_neighbors : int
            number of neighbors to be considered. Must be a positive integer less than n_particles.
        minkowski_p_norm : int
            The Minkowski p-norm to use. 1 is the sum-of-absolute values (or L1 distance) while 2 is the Euclidean
            (or L2) distance.
        ftol : float
            Relative error in objective_func(best_pos) acceptable for convergence.
        initializer : non_linear.initializer.Initializer
            Generates the initialize samples of non-linear parameter space (see autofit.non_linear.initializer).
        number_of_cores : int
            The number of cores Emcee sampling is performed using a Python multiprocessing Pool instance. If 1, a
            pool instance is not created and the job runs in serial.

        All remaining attributes are emcee parameters and described at the PySwarms API webpage:

        https://pyswarms.readthedocs.io/en/latest/index.html
        """

        self.number_of_k_neighbors = (
            self._config("search", "number_of_k_neighbors", int)
            if number_of_k_neighbors is None
            else number_of_k_neighbors
        )

        self.minkowski_p_norm = (
            self._config("search", "minkowski_p_norm", int)
            if minkowski_p_norm is None
            else minkowski_p_norm
        )

        super().__init__(
            paths=paths,
            prior_passer=prior_passer,
            n_particles=n_particles,
            iters=iters,
            cognitive=cognitive,
            social=social,
            inertia=inertia,
            ftol=ftol,
            initializer=initializer,
            iterations_per_update=iterations_per_update,
            number_of_cores=number_of_cores,
        )

        logger.debug("Creating PySwarms NLO")

    def sampler_fom_model_and_fitness(self, model, fitness_function, bounds, init_pos):
        """Get the static Dynesty sampler which performs the non-linear search, passing it all associated input Dynesty
        variables."""

        import pyswarms

        return pyswarms.local_best.LocalBestPSO(
            n_particles=self.n_particles,
            dimensions=model.prior_count,
            bounds=bounds,
            options={
                "c1": self.cognitive,
                "c2": self.social,
                "w": self.inertia,
                "k": self.number_of_k_neighbors,
                "p": self.minkowski_p_norm,
            },
            ftol=self.ftol,
            init_pos=init_pos,
        )



from autofit.mapper.prior.arithmetic import ArithmeticMixin
from autofit.mapper.prior_model.prior_model import PriorModel, Prior


class AnnotationPriorModel(PriorModel, ArithmeticMixin):
    def __init__(self, cls, parent_class, true_argument_name, **kwargs):
        self.parent_class = parent_class
        self.true_argument_name = true_argument_name
        self._value = None
        super().__init__(cls, **kwargs)

    def make_prior(self, attribute_name):
        if self._value is None:
            self._value = Prior.for_class_and_attribute_name(
                self.parent_class, self.true_argument_name
            )
        return self._value



import os

import pytest

from autoconf import conf
import autofit as af
from autofit.non_linear import abstract_search
from test_autofit import mock
from test_autofit import mock_real

directory = os.path.dirname(os.path.realpath(__file__))


class NLO(abstract_search.NonLinearSearch):

    @property
    def config_type(self):
        return conf.instance.mock

    @property
    def tag(self):
        return"nlo"

    def _fit(self, model, analysis):
        class Fitness:
            def __init__(self, instance_from_vector, instance):
                self.result = None
                self.instance_from_vector = instance_from_vector
                self.instance = instance

            def __call__(self, vector):
                instance = self.instance_from_vector(vector)
                for key, value in self.instance.__dict__.items():
                    setattr(instance, key, value)

                log_likelihood = analysis.log_likelihood_function(instance)
                self.result = abstract_search.Result(
                    instance, log_likelihood
                )

                # Return Chi squared
                return -2 * log_likelihood

        fitness_function = Fitness(
            model.instance_from_vector, model.instance_from_prior_medians
        )
        fitness_function(model.prior_count * [0.5])

        return fitness_function.result


@pytest.fixture(name="phase")
def make_phase():
    return MyPhase(af.Paths(name=""), search=NLO)


class MyPhase(af.AbstractPhase):
    prop = af.PhaseProperty("prop")


@pytest.fixture(name="list_phase")
def make_list_phase():
    return MyPhase(af.Paths(name=""), search=NLO)


class TestPhasePropertyList:
    def test_classes(self, list_phase):
        objects = [mock_real.GalaxyModel(), mock_real.GalaxyModel()]

        list_phase.prop = objects

        assert list_phase.model.prop == objects
        assert list_phase.prop == objects

    def test_abstract_prior_models(self, list_phase):
        objects = [af.AbstractPriorModel(), af.AbstractPriorModel()]

        list_phase.prop = objects

        assert list_phase.model.prop == objects
        assert list_phase.prop == objects

    def test_mix(self, list_phase):
        objects = [mock_real.GalaxyModel(), mock.MockComponents()]

        list_phase.prop = objects

        assert list_phase.model.prop == objects
        assert list_phase.prop == objects

    def test_set_item(self, list_phase):
        galaxy_prior_0 = mock_real.GalaxyModel()
        objects = [galaxy_prior_0, mock.MockComponents()]

        list_phase.prop = objects

        galaxy_prior_1 = mock_real.GalaxyModel()
        list_phase.prop[1] = galaxy_prior_1

        assert list_phase.model.prop == [galaxy_prior_0, galaxy_prior_1]

        galaxy = mock.MockComponents()

        list_phase.prop[0] = galaxy
        assert list_phase.prop == [galaxy, galaxy_prior_1]


class TestPhasePropertyCollectionAttributes:
    def test_set_list_as_dict(self, list_phase):
        galaxy_model = mock_real.GalaxyModel()
        list_phase.prop = dict(one=galaxy_model)

        assert len(list_phase.prop) == 1
        # noinspection PyUnresolvedReferences
        assert list_phase.prop.one == galaxy_model

    def test_override_property(self, list_phase):
        galaxy_model = mock_real.GalaxyModel()

        list_phase.prop = dict(one=mock_real.GalaxyModel())

        list_phase.prop.one = galaxy_model

        assert len(list_phase.prop) == 1
        assert list_phase.prop.one == galaxy_model

    def test_named_list_items(self, list_phase):
        galaxy_model = mock_real.GalaxyModel()
        list_phase.prop = [galaxy_model]

        # noinspection PyUnresolvedReferences
        assert getattr(list_phase.prop, "0") == galaxy_model

    def test_mix(self, list_phase):
        objects = dict(one=mock_real.GalaxyModel(), two=mock.MockComponents())

        list_phase.prop = objects

        list_phase.prop.one = mock.MockComponents()

        assert len(list_phase.model.prop) == 2

    def test_named_attributes_in_model(self, list_phase):
        galaxy_model = mock_real.GalaxyModel(model_redshift=True)
        list_phase.prop = dict(one=galaxy_model)

        assert list_phase.model.prior_count == 1
        assert list_phase.model.prop.one == galaxy_model

        instance = list_phase.model.instance_from_prior_medians()

        assert instance.prop.one is not None
        assert len(instance.prop) == 1

    def test_named_attributes_in_model_override(self, list_phase):
        list_phase.prop = dict(one=mock_real.GalaxyModel())

        assert list_phase.model.prior_count == 0

        galaxy_model = mock_real.GalaxyModel(model_redshift=True)

        list_phase.prop.one = galaxy_model

        assert list_phase.model.prior_count == 1
        # assert list_phase.model.one == galaxy_model

        instance = list_phase.model.instance_from_prior_medians()

        # assert instance.one is not None
        assert len(instance.prop) == 1

    def test_named_attributes_in_instance(self, list_phase):
        galaxy = mock.MockComponents()
        list_phase.prop = dict(one=galaxy)

        assert list_phase.model.prior_count == 0
        assert list_phase.model.prop.one == galaxy

    def test_shared_priors(self, list_phase):
        list_phase.prop = dict(
            one=mock_real.GalaxyModel(model_redshift=True), two=mock_real.GalaxyModel(model_redshift=True)
        )

        assert list_phase.model.prior_count == 2

        # noinspection PyUnresolvedReferences
        list_phase.prop.one.redshift = list_phase.prop.two.redshift

        assert list_phase.model.prior_count == 1

    def test_hasattr(self, list_phase):
        list_phase.prop = dict()

        assert not hasattr(list_phase.prop, "one")
        list_phase.prop = dict(one=mock_real.GalaxyModel(model_redshift=True))

        assert hasattr(list_phase.prop, "one")

    def test_position_not_a_prior(self, list_phase):
        list_phase.prop = [af.PriorModel(mock.MockComponents)]

        assert list_phase.model.prior_count == 1
        assert "parameter" == list_phase.model.prior_tuples_ordered_by_id[0][0]

        prior_model = af.PriorModel(mock.MockComponents)
        prior_model.phase_property_position = 0

        print(prior_model.instance_tuples)
        assert len(prior_model.instance_tuples) == 0



import configparser
import logging

from autoconf import conf

logger = logging.getLogger(__name__)


class TextFormatter:
    def __init__(self, line_length=90, indent=4):
        self.dict = dict()
        self.line_length = line_length
        self.indent = indent

    def add_to_dict(self, path_item_tuple: tuple, info_dict: dict):
        path_tuple = path_item_tuple[0]
        key = path_tuple[0]
        if len(path_tuple) == 1:
            info_dict[key] = path_item_tuple[1]
        else:
            if key not in info_dict:
                info_dict[key] = dict()
            self.add_to_dict(
                (path_item_tuple[0][1:], path_item_tuple[1]), info_dict[key]
            )

    def add(self, path_item_tuple: tuple):
        self.add_to_dict(path_item_tuple, self.dict)

    def dict_to_list(self, info_dict, line_length):
        lines = []
        for key, value in info_dict.items():
            indent_string = self.indent * " "
            if isinstance(value, dict):
                sub_lines = self.dict_to_list(
                    value, line_length=line_length - self.indent
                )
                lines.append(key)
                for line in sub_lines:
                    lines.append(f"{indent_string}{line}")
            else:
                value_string = str(value)
                space_string = max((line_length - len(key)), 1) * " "
                lines.append(f"{key}{space_string}{value_string}")
        return lines

    @property
    def list(self):
        return self.dict_to_list(self.dict, line_length=self.line_length)

    @property
    def text(self):
        return "\n".join(self.list)


def format_string_for_parameter_name(parameter_name: str) -> str:
    """
    Get the format for the label. Attempts to extract the key string associated with
    the dimension. Seems dodgy.

    Parameters
    ----------
    parameter_name
        A string label

    Returns
    -------
    format
        The format string (e.g {:.2f})
    """
    label_conf = conf.instance.label_format

    try:
        # noinspection PyProtectedMember
        for key, value in sorted(
            label_conf.parser._sections["format"].items(),
            key=lambda item: len(item[0]),
            reverse=True,
        ):
            if key in parameter_name:
                return value
    except KeyError:
        pass
    raise configparser.NoSectionError(
        "Could not find an entry for the parameter {} in the label_format.iniconfig at path {}".format(
            parameter_name, label_conf.path
        )
    )


def convert_name_to_label(parameter_name, name_to_label):

    if not name_to_label:
        return parameter_name

    label_conf = conf.instance.label

    try:
        return conf.instance.label.label(name=parameter_name)
    except KeyError:
        raise configparser.NoSectionError(
            "Could not find an entry for the parameter {} in the label_format.iniconfig at path {}".format(
                parameter_name, label_conf.path
            )
        )


def add_whitespace(str0, str1, whitespace):
    return f"{str0}{str1.rjust(whitespace - len(str0) + len(str1))}"


def value_result_string_from(
    parameter_name, value, values_at_sigma=None, unit=None, format_string=None
):

    format_str = format_string or format_string_for_parameter_name(parameter_name)
    value = format_str.format(value)

    if unit is not None:
        unit = f" {unit}"
    else:
        unit = ""

    if values_at_sigma is None:
        return f"{value}{unit}"
    else:
        lower_value_at_sigma = format_str.format(values_at_sigma[0])
        upper_value_at_sigma = format_str.format(values_at_sigma[1])
        return f"{value} ({lower_value_at_sigma}, {upper_value_at_sigma}){unit}"


def parameter_result_latex_from(
    parameter_name,
    value,
    errors=None,
    subscript=None,
    unit=None,
    format_string=None,
    name_to_label=False,
):

    format_str = format_string or format_string_for_parameter_name(parameter_name)
    value = format_str.format(value)

    str0 = convert_name_to_label(
        parameter_name=parameter_name, name_to_label=name_to_label
    )

    if unit is not None:
        unit = f" {unit}"
    else:
        unit = ""

    if subscript is None:
        subscript = ""
    else:
        subscript = f"_{{\mathrm{{{subscript}}}}}"

    if errors is None:
        return f"{str0}{subscript} = {value}{unit} & "
    else:
        lower_value_at_sigma = format_str.format(errors[0])
        upper_value_at_sigma = format_str.format(errors[1])
        return f"{str0}{subscript} = {value}^{{+{upper_value_at_sigma}}}_{{-{lower_value_at_sigma}}}{unit} & "


def output_list_of_strings_to_file(file, list_of_strings):
    file = open(file, "w")
    file.write("".join(list_of_strings))
    file.close()



#!/usr/bin/env python

"""
Filter and collate phase dataset in all subdirectories.

Usage:

./aggregator.py (root_directory) [pipeline=pipeline phase=phase dataset=dataset]

Example:

./aggregator.py ../output pipeline=data_mass_x1_source_x1_positions
"""

import os
import zipfile
from collections import defaultdict
from shutil import rmtree
from typing import List, Union, Iterator, Tuple

from .phase_output import PhaseOutput
from .predicate import AttributePredicate


class AggregatorGroup:
    def __init__(self, groups: ["AbstractAggregator"]):
        """
        A group of aggregators produced by grouping phases on a field.

        Parameters
        ----------
        groups
            Groups, each with a common value in the metadata file
        """
        self.groups = groups

    def filter(self, *predicates) -> "AggregatorGroup":
        """
        Apply predicates to the underlying groups whilst maintaining the total number of groups.

        Parameters
        ----------
        predicates
            Predicates that evaluate to True or False for any given phase.

        Returns
        -------
        A collection of groups of the same length with each group having the same or fewer members.
        """
        return AggregatorGroup(
            [
                group.filter(*predicates)
                for group in self.groups
            ]
        )

    def __getitem__(self, item):
        return self.groups[item]

    def __len__(self):
        return len(self.groups)

    def values(self, name: str) -> List[List]:
        """
        Extract a list of lists values with a certain name from the output objects in
        this group.

        Parameters
        ----------
        name
            The name of the attribute to be extracted

        Returns
        -------
        A list of lists of values.
        """
        return [
            group.values(name)
            for group in self.groups
        ]


class AbstractAggregator:
    def __init__(self, phases: List[PhaseOutput]):
        """
        An aggregator that comprises several phases which matching filters.

        Parameters
        ----------
        phases
            Phases that were found to have matching filters
        """
        self.phases = phases

    def remove_unzipped(self):
        """
        Removes the unzipped output directory for each phase.
        """
        for phase in self.phases:
            path = "/".join(
                phase.directory.split("/")[:-1]
            )
            rmtree(
                path,
                ignore_errors=True
            )

    def __getitem__(
            self,
            item: Union[slice, int]
    ) -> Union[
        "AbstractAggregator",
        PhaseOutput
    ]:
        """
        If an index is passed in then a specific phase output is returned.

        If a slice is passed in then an aggregator comprising several phases is returned.

        Parameters
        ----------
        item
            A slice or index

        Returns
        -------
        An aggregator or phase
        """
        if isinstance(item, slice):
            return AbstractAggregator(
                self.phases[item]
            )
        return self.phases[item]

    def __len__(self):
        return len(self.phases)

    def __getattr__(self, item):
        return AttributePredicate(item)

    def filter(self, *predicates) -> "AbstractAggregator":
        """
        Filter phase outputs by predicates. A predicate is created using a conditional
        operator.

        Another aggregator object is returned.

        Parameters
        ----------
        predicates
            Objects representing predicates that may evaluate to True or False for any
            given phase output.

        Returns
        -------
        An aggregator comprising all phases that evaluate to True for all predicates.
        """
        phases = self.phases
        for predicate in predicates:
            phases = predicate.filter(phases)
        phases = list(phases)
        print(f"Filter found a total of {str(len(phases))} results")
        return AbstractAggregator(phases=list(phases))

    def values(self, name: str) -> Iterator:
        """
        Get values from outputs with a given name.

        A list the same length as the number of phases is returned
        where each item is the value of the attribute for a given
        phase.

        Parameters
        ----------
        name
            The name of an attribute expected to be associated with
            phase output. If a pickle file with this name is in the
            phase output directory then that pickle will be loaded.

        Returns
        -------
        A generator of values for the attribute
        """
        return map(
            lambda phase: getattr(
                phase, name
            ),
            self.phases
        )

    def homogenize(
            self,
            aggregator: "AbstractAggregator",
            on: str
    ) -> Tuple[
        "AbstractAggregator",
        "AbstractAggregator"
    ]:
        """
        Filter this aggregator and another aggregator such that each only
        contain results where at least one result in the other aggregator
        has the same value for a given property.

        Parameters
        ----------
        aggregator
            Another aggregator to homogenize with this one.
        on
            The attribute of the underlying results which should match.

        Returns
        -------
        A pair of aggregators with only matching results.
        """
        def _homogenize(a, b):
            values = set(b.values(on))
            return AbstractAggregator([
                phase for phase in a.phases
                if getattr(phase, on) in values
            ])

        return _homogenize(
            self,
            aggregator
        ), _homogenize(
            aggregator,
            self
        )

    def map(self, func):
        """
        Map some function onto the aggregated output objects.

        Parameters
        ----------
        func
            A function

        Returns
        -------
        A generator of results
        """
        return map(
            func,
            self.phases
        )

    def group_by(self, field: str) -> AggregatorGroup:
        """
        Group the phases by a field, e.g. pipeline.

        The object returned still permits filtering and attribute querying.

        Parameters
        ----------
        field
            The field by which to group

        Returns
        -------
        An object comprising lists of grouped fields
        """
        group_dict = defaultdict(list)
        for phase in self.phases:
            group_dict[getattr(phase, field)].append(phase)
        return AggregatorGroup(list(map(AbstractAggregator, group_dict.values())))

    @property
    def model_results(self) -> str:
        """
        A string joining headers and results for all included phases.
        """
        return "\n\n".join(
            "{}\n\n{}".format(phase.header, phase.model_results)
            for phase in self.phases
        )


class Aggregator(AbstractAggregator):
    def __init__(
            self,
            directory: str,
            completed_only=False
    ):
        """
        Class to aggregate phase results for all subdirectories in a given directory.

        The whole directory structure is traversed and a Phase object created for each directory that contains a
        metadata file.

        Parameters
        ----------
        directory
            A directory in which the outputs of phases are kept. This is searched recursively.
        completed_only
            If True only phases with a .completed file (indicating the phase was completed)
            are included in the aggregator.
        """

        # TODO : Progress bar here

        print("Aggregator loading phases... could take some time.")

        self._directory = directory
        phases = []

        for root, _, filenames in os.walk(directory):
            for filename in filenames:
                if filename.endswith(".zip"):
                    with zipfile.ZipFile(f"{root}/{filename}", "r") as f:
                        f.extractall(f"{root}/{filename[:-4]}")

        for root, _, filenames in os.walk(directory):
            if "metadata" in filenames:
                if not completed_only or ".completed" in filenames:
                    phases.append(PhaseOutput(root))

        if len(phases) == 0:
            print(f"\nNo phases found in {directory}\n")
        else:
            print(f"\n A total of {str(len(phases))} phases and results were found.")
        super().__init__(phases)



class TestFiltering:
    def test_or(self, aggregator):
        predicate_one = aggregator.directory.contains(
            "one"
        )
        predicate_two = aggregator.directory.contains(
            "two"
        )
        result = aggregator.filter(
            predicate_one | predicate_two
        )
        assert len(result) == 2
        assert result.directories == [
            "directory/number/one",
            "directory/number/two"
        ]

    def test_and(self, aggregator):
        predicate_one = aggregator.pipeline == "pipeline1"
        predicate_two = aggregator.phase == "phase2"
        result = aggregator.filter(
            predicate_one & predicate_two
        )
        assert len(result) == 1
        assert result.directories == [
            "directory/number/two"
        ]

    def test_not_contains(self, aggregator):
        predicate = ~(aggregator.pipeline.contains("1"))
        result = aggregator.filter(
            predicate
        )
        assert len(result) == 1
        assert result[0].pipeline == "pipeline2"

    def test_not_equal(self, aggregator):
        predicate = aggregator.pipeline != "pipeline1"
        result = aggregator.filter(
            predicate
        )
        assert len(result) == 1
        assert result[0].pipeline == "pipeline2"

    def test_rhs(self, aggregator):
        predicate = "pipeline1" != aggregator.pipeline
        result = aggregator.filter(
            predicate
        )
        assert result.pipeline == ["pipeline2"]

        predicate = "pipeline1" == aggregator.pipeline
        result = aggregator.filter(
            predicate
        )
        assert result.pipeline == ["pipeline1", "pipeline1"]

    def test_filter_index(self, aggregator):
        assert list(aggregator.filter(
            aggregator.pipeline == "pipeline1"
        )[1:].values("pipeline")) == ["pipeline1"]
        assert list(aggregator[1:].filter(
            aggregator.pipeline == "pipeline1"
        ).values("pipeline")) == ["pipeline1"]

    def test_filter(self, aggregator):
        result = aggregator.filter(
            aggregator.pipeline == "pipeline1"
        )
        assert len(result) == 2
        assert result[0].pipeline == "pipeline1"

        result = aggregator.filter(
            aggregator.pipeline == "pipeline1",
            aggregator.phase == "phase1"
        )

        assert len(result) == 1
        assert result[0].pipeline == "pipeline1"

        result = aggregator.filter(
            aggregator.pipeline == "pipeline1"
        ).filter(
            aggregator.phase == "phase1"
        )

        assert len(result) == 1
        assert result[0].pipeline == "pipeline1"

    def test_filter_contains(self, aggregator):
        result = aggregator.filter(
            aggregator.pipeline.contains("1")
        )
        assert len(result) == 2
        assert result[0].pipeline == "pipeline1"

        result = aggregator.filter(
            aggregator.pipeline.contains("1"),
            aggregator.phase.contains("1")
        )

        assert len(result) == 1
        assert result[0].pipeline == "pipeline1"

        result = aggregator.filter(
            aggregator.pipeline.contains("1")
        ).filter(
            aggregator.phase == "phase1"
        )

        assert len(result) == 1
        assert result[0].pipeline == "pipeline1"

    def test_filter_contains_directory(self, aggregator):
        result = aggregator.filter(
            aggregator.directory.contains("number")
        )
        assert len(result) == 2

        result = aggregator.filter(
            aggregator.directory.contains("letter")
        )
        assert len(result) == 1



from itertools import repeat
from typing import Optional, Dict, Tuple, NamedTuple, Any

import numpy as np
from scipy.optimize import minimize, OptimizeResult, least_squares

from autofit.graphical import FixedMessage
from autofit.mapper.variable import Variable
from autofit.graphical.factor_graphs import Factor
from .mean_field import FactorApproximation, MeanFieldApproximation, Status
from .utils import propagate_uncertainty, FlattenArrays


class OptResult(NamedTuple):
    mode: Dict[str, np.ndarray]
    inv_hessian: Dict[str, np.ndarray]
    result: OptimizeResult


class OptFactor:
    """
    """

    def __init__(
            self,
            factor_approx: FactorApproximation,
            kwargs: Dict[Variable, Tuple[int, ...]],
            fixed_kws: Optional[Dict[str, np.ndarray]] = None,
            bounds: Optional[Dict[str, Tuple[float, float]]] = None,
            sign: int = 1, method: str = 'L-BFGS-B',
    ):
        self.factor_approx = factor_approx
        self.param_shapes = FlattenArrays(kwargs)
        self.param_bounds = bounds
        self.free_vars = tuple(kwargs)

        self.sign = sign
        self.fixed_kws = fixed_kws
        self.method = method

        if bounds:
            # TODO check that this is correct for composite
            # distributions e.g. NormalGammaMessage
            self.bounds = [
                b for k, s in kwargs.items()
                for bound in bounds[k]
                for b in repeat(bound, np.prod(s, dtype=int))]
        else:
            self.bounds = bounds

    @classmethod
    def from_approx(
            cls,
            factor_approx: FactorApproximation,
            **kwargs
    ) -> 'OptFactor':
        fixed_kws = {}
        bounds = {}
        for v in factor_approx.factor.variables:
            dist = factor_approx.model_dist[v]
            if isinstance(dist, FixedMessage):
                fixed_kws[v] = dist.mean
            else:
                kwargs[v] = dist.shape
                bounds[v] = dist._support

        return cls(
            factor_approx,
            kwargs,
            fixed_kws=fixed_kws,
            sign=-1,
            bounds=bounds,
        )

    def __call__(self, args):
        params = self.param_shapes.unflatten(args)
        params.update(self.fixed_kws)
        return self.sign * np.sum(self.factor_approx(params))

    def _minimise(self, arrays_dict, method=None, bounds=None,
                  constraints=(), tol=None, callback=None,
                  options=None):
        x0 = self.param_shapes.flatten(arrays_dict)
        bounds = self.bounds if bounds is None else bounds
        method = self.method if method is None else method
        return minimize(
            self, x0, method=method, bounds=bounds,
            constraints=constraints, tol=tol, callback=callback,
            options=options)

    def _parse_result(self, res: OptimizeResult) -> OptResult:
        return OptResult(
            self.param_shapes.unflatten(res.x),
            self.param_shapes.unflatten(res.hess_inv.todense()),
            res)

    def minimise(self, bounds=None,
                 constraints=(), tol=None, callback=None,
                 options=None, **arrays):
        self.sign = 1
        res = self._minimise(
            bounds=bounds, constraints=constraints, tol=tol,
            callback=callback, options=options, **arrays)
        return self._parse_result(res)

    def maximise(
            self,
            arrays_dict,
            bounds=None,
            constraints=(), tol=None, callback=None,
            options=None
    ):
        self.sign = -1
        res = self._minimise(
            arrays_dict,
            bounds=bounds, constraints=constraints, tol=tol,
            callback=callback, options=options)
        self.sign = 1
        return self._parse_result(res)

    minimize = minimise
    maximize = maximise


def maximise_factor_approx(
        factor_approx: FactorApproximation, **kwargs):
    """
    """
    p0 = {
        v: kwargs.pop(v, factor_approx.model_dist[v].sample(1)[0])
        for v in factor_approx.factor.variables}
    opt = OptFactor.from_approx(factor_approx, **kwargs)
    return opt.maximise(**p0)


maximize_factor_approx = maximise_factor_approx


def find_factor_mode(
        factor_approx: FactorApproximation,
        return_cov: bool = True,
        status: Optional[Status] = None,
        min_iter: int = 2,
        **kwargs
):
    """
    """
    success, messages = Status() if status is None else status
    factor = factor_approx.factor

    opt = OptFactor.from_approx(factor_approx, **kwargs)
    free_vars = opt.free_vars
    det_vars = factor_approx.deterministic_dist
    p0 = {
        v: kwargs.pop(v, factor_approx.model_dist[v].sample(1)[0])
        for v in free_vars}

    # find the mode of the factor approximation
    res = opt.maximise(p0)
    mode = {**res.mode, **opt.fixed_kws}

    result = res.result
    success = result.success
    message = result.message.decode()
    messages += (
        "optimise.find_factor_mode: "
        f"nfev={result.nfev}, nit={result.nit}, "
        f"status={result.status}, message={message}",)
    if result.nit < min_iter:
        success = False
        # TODO implement numerical Hessian

    # find the values of the deterministic variables at the mode
    f = factor(mode)
    mode.update(f.deterministic_values)

    if return_cov:
        # Calculate covariance of deterministic variables based on
        # the free variables covariance and Jacobians
        covars = res.inv_hessian
        factor_jac = factor.jacobian(free_vars, mode)
        for det in det_vars:
            covars[det] = 0.
            for v in free_vars:
                cov = covars[v]
                jac = factor_jac.deterministic_values[det, v]
                det_cov = propagate_uncertainty(cov, jac)
                covars[det] += det_cov
    else:
        covars = {}

    return mode, covars, Status(success, messages), result


class LaplaceOptimiser:
    def __init__(
            self,
            n_iter=4,
            delta=1.
    ):
        self.history = dict()
        self.n_iter = n_iter
        self.delta = delta

    def run(self, model_approx):
        model = model_approx.factor_graph
        for i in range(self.n_iter):
            for factor in model.factors:
                # We have reduced the entire EP step into a single function
                model_approx, status = self.laplace_factor_approx(
                    model_approx,
                    factor
                )

                # save and print current approximation
                self.history[i, factor] = model_approx
        return model_approx

    def laplace_factor_approx(
            self,
            model_approx,
            factor: Factor,
            opt_kws: Optional[Dict[str, Any]] = None
    ):
        factor_approx = model_approx.factor_approximation(factor)

        opt_kws = {} if opt_kws is None else opt_kws
        mode, covars, status, _ = find_factor_mode(
            factor_approx,
            return_cov=True,
            **opt_kws
        )

        model_dist = {
            v: factor_approx.factor_dist[v].from_mode(
                mode[v],
                covars.get(v)
            )
            for v in mode
        }

        projection, status = factor_approx.project(
            model_dist,
            delta=self.delta,
            status=status
        )

        return model_approx.project(
            projection,
            status=status
        )


class LeastSquaresOpt:
    _opt_params = dict(
        jac='2-point', method='trf', ftol=1e-08,
        xtol=1e-08, gtol=1e-08, x_scale=1.0, loss='linear',
        f_scale=1.0, diff_step=None, tr_solver=None,
        tr_options={}, jac_sparsity=None, max_nfev=None,
        verbose=0)

    def __init__(
            self,
            factor_approx: FactorApproximation,
            fixed_kws=None,
            param_bounds=None,
            opt_only=None,
            **kwargs):

        self.factor_approx = factor_approx
        self.opt_params = {**self._opt_params, **kwargs}

        param_shapes = {}
        param_bounds = {} if param_bounds is None else param_bounds
        fixed_kws = {} if fixed_kws is None else fixed_kws

        for v in factor_approx.factor.variables:
            dist = factor_approx.model_dist[v]
            if isinstance(dist, FixedMessage):
                fixed_kws[v] = dist.mean
            else:
                param_shapes[v] = dist.shape
                param_bounds[v] = dist._support

        self.fixed_kws = fixed_kws
        self.param_shapes = FlattenArrays(param_shapes)

        if opt_only is None:
            opt_only = tuple(
                v for v, d in factor_approx.all_cavity_dist.items()
                if not isinstance(d, FixedMessage)
            )

        self.opt_only = opt_only
        self.resid_means = {
            k: factor_approx.all_cavity_dist[k].mean
            for k in self.opt_only}
        self.resid_scales = {
            k: factor_approx.all_cavity_dist[k].scale
            for k in self.opt_only}
        self.resid_shapes = FlattenArrays({
            k: np.shape(m) for k, m in
            self.resid_means.items()})

        self.bounds = tuple(np.array(list(zip(*[
            b for k, s in param_shapes.items()
            for bound in param_bounds[k]
            for b in repeat(bound, np.prod(s, dtype=int))]))))

    def __call__(self, arr):
        p0 = self.param_shapes.unflatten(arr)
        log_value, det_vars = self.factor_approx.factor(
            {**p0, **self.fixed_kws}
        )
        vals = {**p0, **det_vars}
        residuals = {
            v: (vals[v] - mean) / self.resid_scales[v]
            for v, mean in self.resid_means.items()
        }
        return self.resid_shapes.flatten(residuals)

    def least_squares(self):
        p0 = {
            v: self.factor_approx.model_dist[v].sample(1)[0]
            for v in self.param_shapes.keys()}
        arr = self.param_shapes.flatten(p0)

        res = least_squares(
            self, arr, bounds=self.bounds, **self.opt_params)

        sol = self.param_shapes.unflatten(res.x)
        _, det_vars = self.factor_approx.factor(
            {**sol, **self.fixed_kws}
        )

        jac = {
            (d, k): b
            for k, a in self.param_shapes.unflatten(
                res.jac.T, ndim=1).items()
            for d, b in self.resid_shapes.unflatten(
                a.T, ndim=1).items()}
        hess = self.param_shapes.unflatten(
            res.jac.T.dot(res.jac))

        def inv(a):
            shape = np.shape(a)
            ndim = len(shape)
            if ndim:
                a = np.asanyarray(a)
                s = shape[:ndim // 2]
                n = np.prod(s, dtype=int)
                return np.linalg.inv(
                    a.reshape(n, n)).reshape(s + s)
            else:
                return 1 / a

        invhess = {
            k: inv(h) for k, h in hess.items()}
        for det in det_vars:
            invhess[det] = 0.
            for v in sol:
                invhess[det] += propagate_uncertainty(
                    invhess[v], jac[det, v])

        mode = {**sol, **det_vars}
        return mode, invhess, res


def lstsq_laplace_factor_approx(
        model_approx: MeanFieldApproximation,
        factor: Factor,
        delta: float = 0.5,
        opt_kws: Optional[Dict[str, Any]] = None):
    """
    """
    factor_approx = model_approx.factor_approximation(factor)

    opt = LeastSquaresOpt(
        factor_approx, **({} if opt_kws is None else opt_kws))

    mode, covar, result = opt.least_squares()
    message = (
        "optimise.lsq_sq_laplace_factor_approx: "
        f"nfev={result.nfev}, njev={result.njev}, "
        f"optimality={result.optimality}, "
        f"cost={result.cost}, "
        f"status={result.status}, message={result.message}",)
    status = Status(result.success, message)

    model_dist = {
        v: factor_approx.factor_dist[v].from_mode(
            mode[v],
            covar.get(v))
        for v in mode
    }

    projection, status = factor_approx.project(
        model_dist, delta=delta, status=status)

    return model_approx.project(projection, status=status)



import autofit as af
from howtofit.chapter_1_introduction.tutorial_4_source_code.src.dataset.dataset import (
    Dataset,
)
from howtofit.chapter_1_introduction.tutorial_4_source_code.src.phase.result import (
    Result,
)
from howtofit.chapter_1_introduction.tutorial_4_source_code.src.phase.analysis import (
    Analysis,
)

"""
The phase package combines a data-set, model and `NonLinearSearch`, allowing us to fit the `Dataset` with the model. It
essentially acts as the `meeting point` between the other packages in the project (dataset, fit, plot) and modules
in the phase package (phase.py, analysis.py, result.py).
"""


class Phase(af.AbstractPhase):

    """
    This tells the phase that the input parameter `gaussian` is a model component that is fitted for by the phase`s
    `NonLinearSearch`.

    In `analysis.py`, the function `fit` has an input parameter called `instance` which is the gaussian mapped from
    this model via a unit vector and the model priors (as described in tutorial 1).

    For your model-fitting problem, this will be replaced by the modules in your `model` package.
    """

    gaussian = af.PhaseProperty("gaussian")

    Result = Result  # Set the result to the Result class in `result.py`

    @af.convert_paths  # <- This handles setting up output paths.
    def __init__(
        self,
        paths,
        gaussian,  # <- The user inputs a model -> gaussian.py -> `Gaussian` class here.
        search,  # <- This specifies the default `NonLinearSearch` used by the phase.
    ):
        """
        A phase which fits a `Gaussian` model using a `NonLinearSearch`.

        Parameters
        ----------
        paths : af.Paths
            Handles the output directory structure.
        gaussian : model.gaussians.Gaussian
            The model component `Gaussian` class fitted by this phase.
        search: class
            The class of a non_linear search
        """
        super().__init__(paths=paths, search=search)
        self.gaussian = gaussian

    def run(self, dataset: Dataset):
        """ Pass a `Dataset` to the phase, running the phase and `NonLinearSearch`.

        Parameters
        ----------
        `Dataset` : `Dataset`.Dataset
            The `Dataset` fitted by the phase, which is specified in the module `dataset/dataset.py`

        Returns
        -------
        result: result.Result
            A result object comprising information on the `NonLinearSearch` and the maximum likelihood model.
        """

        """
        These functions create instances of the Analysis class (in `analysis.py`), runs the analysis (which performs
        the `NonLinearSearch` ) and returns an instance of the Result class (in `result.py`).

        Once you`ve looked through this module, check those modules out to see exactly what these classes do!
        """

        analysis = self.make_analysis(dataset=dataset)

        """
        `run_analysis` is not located in analysis.py, instead it is an inherited method from the parent class
        `af.AbstractPhase`. Essentially, all this function does is begin the `NonLinearSearch`, using the analysis
        created above.
        """

        result = self.run_analysis(analysis=analysis)

        return self.make_result(result=result, analysis=analysis)

    def make_analysis(self, dataset):
        """
        Create an Analysis object, which uses the `Dataset` with functions to perform a fit.

        Parameters
        ----------
        `Dataset` : `Dataset`.Dataset
            The `Dataset` fitted by the phase, which is specified in the module `dataset/dataset.py`

        Returns
        -------
        analysis : Analysis
            An analysis object that the `NonLinearSearch` calls to determine the fit log_likelihood for a given model
            instance.
        """
        return Analysis(dataset=dataset)

    def make_result(self, result, analysis):
        return self.Result(
            samples=result.samples,
            previous_model=self.model,
            search=self.search,
            analysis=analysis,
        )



import numpy as np
import pytest
from scipy import stats

import autofit.graphical.messages.normal
from autofit import graphical as mp


@pytest.fixture(
    name="normal_factor"
)
def make_normal_factor(x):
    return mp.Factor(
        stats.norm(
            loc=-0.5,
            scale=0.5
        ).logpdf,
        x=x
    )


@pytest.fixture(
    name="model"
)
def make_model(
        probit_factor,
        normal_factor
):
    return probit_factor * normal_factor


@pytest.fixture(
    name="model_approx"
)
def make_model_approx(
        model,
        x
):
    return mp.MeanFieldApproximation.from_kws(
        model,
        {x: autofit.graphical.messages.normal.NormalMessage(0, 1)}
    )


@pytest.fixture(
    name="probit_approx"
)
def make_probit_approx(
        probit_factor,
        model_approx
):
    return model_approx.factor_approximation(
        probit_factor
    )


def test_approximations(
        probit_approx,
        model_approx,
        x
):
    opt_probit = mp.OptFactor.from_approx(probit_approx)
    result = opt_probit.maximise({x: 0.})

    probit_model = autofit.graphical.messages.normal.NormalMessage.from_mode(
        result.mode[x],
        covariance=result.inv_hessian[x]
    )

    probit_model_dist = {x: probit_model}

    # get updated factor approximation
    probit_project, status = probit_approx.project(
        probit_model_dist, delta=1.
    )

    assert probit_project.model_dist[x].mu == pytest.approx(0.506, rel=0.1)
    assert probit_project.model_dist[x].sigma == pytest.approx(0.814, rel=0.1)

    assert probit_project.factor_dist[x].mu == pytest.approx(1.499, rel=0.1)
    assert probit_project.factor_dist[x].sigma == pytest.approx(1.401, rel=0.1)


def test_looped_importance_sampling(
        model,
        normal_factor,
        probit_factor,
        x
):
    model_approx = mp.MeanFieldApproximation.from_kws(
        model,
        {x: autofit.graphical.messages.normal.NormalMessage(0, 1)}
    )

    np.random.seed(1)
    sampler = mp.ImportanceSampler(
        n_samples=1000
    )
    history = list()

    for i in range(3):
        for factor in [normal_factor, probit_factor]:
            # get factor, cavity distribution and model approximation
            factor_approx = model_approx.factor_approximation(factor)
            # sample from model approximation
            sample = sampler.sample(factor_approx)
            # project sufficient statistcs of sample onto normal dist
            model_dist = mp.project_factor_approx_sample(
                factor_approx,
                sample
            )

            # divide projection by cavity distribution
            factor_project, status = factor_approx.project(
                model_dist,
                delta=1
            )

            # update model approximation
            model_approx, status = model_approx.project(
                factor_project,
                status
            )

            # save and print current approximation
            history.append(model_approx)
            print(i, factor, model_approx[x])

    result = history[-1][x]

    assert result.mu == pytest.approx(-0.243, rel=0.1)
    assert result.sigma == pytest.approx(0.466, rel=0.1)



from abc import ABC, abstractmethod
from functools import wraps
from typing import List, Tuple, cast, Collection, Set
from itertools import count

import numpy as np

from autofit.mapper.variable import Variable, Plate


def accept_variable_dict(func):
    @wraps(func)
    def wrapper(self, variable_dict=None, **kwargs):
        if variable_dict is not None:
            kwargs.update({
                variable.name if isinstance(
                    variable,
                    Variable
                ) else variable: array
                for variable, array
                in variable_dict.items()
            })
        return func(
            self,
            **kwargs
        )

    return wrapper


class AbstractNode(ABC):
    _deterministic_variables: Set[Variable]
    _id = count()

    def __init__(
            self,
            **kwargs: Variable
    ):
        """
        A node in a factor graph

        Parameters
        ----------
        args
            Positional arguments passed to the factor
        kwargs
            Key word arguments passed to the value
        """
        self._kwargs = kwargs
        self.id = next(self._id)

    @property
    @abstractmethod
    def variables(self):
        pass

    @property
    def variable_names(self):
        return {
            variable.name
            for variable
            in self.variable_names
        }

    @property
    @abstractmethod
    def deterministic_variables(self):
        return self._deterministic_variables

    def __getattr__(self, item):
        try:
            return self._kwargs[
                item
            ]
        except KeyError:
            for variable in self.variables | self._deterministic_variables:
                if variable.name == item:
                    return variable
            raise AttributeError(
                f"No attribute {item}"
            )

    @property
    @abstractmethod
    def name(self) -> str:
        """
        A name for this object
        """

    @property
    def call_signature(self) -> str:
        """
        The apparent signature of this object
        """
        args = ", ".join(self.arg_names)
        kws = ", ".join(map("{0[0]}={0[1]}".format, self.kwarg_names))
        call_strs = []
        if args:
            call_strs.append(args)
        if kws:
            call_strs.extend(['*', kws])
        call_str = ", ".join(call_strs)
        call_sig = f"{self.name}({call_str})"
        return call_sig

    @property
    def kwarg_names(self) -> List[str]:
        """
        The names of the variables passed as keyword arguments
        """
        return [
            arg.name
            for arg
            in self._kwargs.values()
        ]

    @property
    def all_variables(self) -> Set[Variable]:
        """
        A dictionary of variables associated with this node
        """
        return self.variables | self._deterministic_variables

    def _broadcast(
            self,
            plate_inds: np.ndarray,
            value: np.ndarray
    ) -> np.ndarray:
        """
        Ensure the shape of the data matches the shape of the plates

        Parameters
        ----------
        plate_inds
            The indices of the plates of some factor within this node
        value
            Some data

        Returns
        -------
        The data reshaped
        """
        shape = np.shape(value)
        shift = len(shape) - plate_inds.size

        assert shift in {0, 1}
        newshape = np.ones(self.ndim + shift, dtype=int)
        newshape[:shift] = shape[:shift]
        newshape[shift + plate_inds] = shape[shift:]

        return np.reshape(value, newshape)

    @property
    def plates(self) -> Tuple[Plate]:
        """
        A tuple of the set of all plates in this graph, ordered by id
        """
        return tuple(sorted(set(
            cast(Plate, plate)
            for variable
            in self.all_variables
            for plate in variable.plates
        )))

    @property
    def ndim(self) -> int:
        """
        The number of plates contained within this graph's variables

        That is, the total dimensions of those variables.
        """
        return len(self.plates)

    def _match_plates(
            self,
            plates: Collection[Plate]
    ) -> np.ndarray:
        """
        Find indices plates from some factor in the collection of
        plates associated with this node.

        Parameters
        ----------
        plates
            Plates from some other node

        Returns
        -------
        An array of plate indices
        """
        return np.array([self.plates.index(p) for p in plates], dtype=int)

    @abstractmethod
    def __call__(self, **kwargs):
        pass



import autofit as af

from howtofit.chapter_1_introduction.tutorial_6_complex_models.src.fit import (
    fit as f,
)
from howtofit.chapter_1_introduction.tutorial_6_complex_models.src.phase import (
    visualizer,
)

"""
The `analysis.py` module in this tutorial is changed from tutorial 5, such that the fit function now assumes multiple
profiles are passed to it rather than a single Gaussian, see the comments below!
"""


class Analysis(af.Analysis):
    def __init__(self, masked_dataset, image_path=None):

        super().__init__()

        """The `MaskedDataset` and visualizer are created in the same way as tutorial 5."""

        self.masked_dataset = masked_dataset

        self.visualizer = visualizer.Visualizer(
            masked_dataset=self.masked_dataset, image_path=image_path
        )

    def log_likelihood_function(self, instance):
        """Determine the fit of a list of Profiles (Gaussians, Exponentials, etc.) to the dataset, using a
        model instance.

        Parameters
        ----------
        instance
            The list of Profile model instance (e.g. the Gaussians, Exponentials, etc.).

        Returns
        -------
        fit : Fit.log_likelihood
            The log likelihood value indicating how well this model fit the `MaskedDataset`.
        """

        """
        In tutorials 3 & 4, the instance was an instance of a single `Gaussian` profile. PyAutoFit knew this instance
        would contain just one Gaussian, because when the phase was created we used a `PriorModel` object in PyAutoFit
        to make the Gaussian. This meant we could create the model data using the line:

            model_data = instance.gaussian.profile_from_xvalues(xvalues=self.masked_dataset.xvalues)

        In this tutorial our instance is comprised of multiple Profile objects. This is reflected the main tutorial
        script where we create a model using a CollectionPriorModel:

            model = CollectionPriorModel(gaussian=profiles.Gaussian, exponential=profiles.Exponential).

        By using a CollectionPriorModel, this means the instance parameter input into the fit function is a
        dictionary where individual profiles (and their parameters) can be accessed as followed:
        
            print(instance.profiles.gaussian)
            print(instance.profiles.exponential)
            print(instance.profiles.exponential.centre)

        The names of the attributes of the instance correspond to what we input into the CollectionPriorModel. Lets
        look at a second example:

            model = CollectionPriorModel(
                          gaussian_0=profiles.Gaussian,
                          gaussian_1=profiles.Gaussian,
                          whatever_i_want=profiles.Exponential
                     ).

            print(instance.profiles.gaussian_0)
            print(instance.profiles.gaussian_1)
            print(instance.profiles.whatever_i_want.centre)

        A CollectionPriorModel allows us to name our model components whatever we want!

        In this tutorial, we want our `fit` function to fit the data with a profile which is the summed profile
        of all individual profiles in the model. Look at `model_data_from_instance` to see how we do this.
        """

        model_data = self.model_data_from_instance(instance=instance)
        fit = self.fit_from_model_data(model_data=model_data)
        return fit.log_likelihood

    def model_data_from_instance(self, instance):

        """
        To create the summed profile of all individual profiles in an instance, we can use a list comprehension
        to iterate over all profiles in the instance.
        """

        return sum(
            [
                profile.profile_from_xvalues(xvalues=self.masked_dataset.xvalues)
                for profile in instance.profiles
            ]
        )

        """
        For those not familiar with list comprehensions, below I`ve included how one would use the instance to create
        the summed profile using a more simple for loop.
        
            model_data = np.zeros(shape=self.masked_dataset.xvalues.shape[0])

            for profile in instance:
                model_data += profile.profile_from_xvalues(xvalues=self.masked_dataset.xvalues)
    
            return model_data
        """

    def fit_from_model_data(self, model_data):
        return f.FitDataset(masked_dataset=self.masked_dataset, model_data=model_data)

    def visualize(self, instance, during_analysis):

        """
        It is worth noting here why we create specific methods for creating the model_data and fit in an analysis.
        By doing so, the code in our visualize function (and also in the `result.py` module) do not need changing
        even though we are now using a model with more components, requiring to sum their individual profiles.
        """

        model_data = self.model_data_from_instance(instance=instance)
        fit = self.fit_from_model_data(model_data=model_data)

        self.visualizer.visualize_fit(fit=fit, during_analysis=during_analysis)



# %%
"""
Tutorial 6: Complex Models
==========================

Up to now, we've fitted a very simple model, a 1D `Gaussian` with 3 free parameters. In this tutorial, we'll look at
how **PyAutoFit** allows us to compose and fit models of arbitrary complexity.

To begin, you should check out the module `tutorial_6_complex_models/model/profiles.py`. In previous tutorials this
module was called `gaussian.py` and contained only the `Gaussian` class. The module now includes a second profile,
_Exponential_, which like the `Gaussian` class is a model-component that can be fitted to data.

Up to now, our data has always been generated using a single `Gaussian` profile. Thus, we have only needed to fit
it with a single `Gaussian`. In this tutorial, our `Dataset` is now a superpositions of multiple profiles. The models
we compose and fit are therefore composed of multiple profiles, such that when we generate the model-data we
generate it as the sum of all individual profiles in our model.
"""

# %%
#%matplotlib inline

from autoconf import conf
import autofit as af
from howtofit.chapter_1_introduction.tutorial_6_complex_models import (
    src as htf,
)

import numpy as np

import os

workspace_path = os.environ["WORKSPACE"]
print("Workspace Path: ", workspace_path)

# %%
"""
Setup the configs as we did in the previous tutorial, as well as the output folder for our `NonLinearSearch`.
"""

# %%
conf.instance = conf.Config(
    config_path=f"{workspace_path}/config",
    output_path=f"{workspace_path}/output/chapter_1",
)

# %%
"""
Lets quickly recap tutorial 1, where using `PriorModels` we created a `Gaussian` as a model component and used it to 
map a list of parameters to a model `instance`.
"""

# %%
model = af.PriorModel(htf.profiles.Gaussian)

print("PriorModel `Gaussian` object: \n")
print(model)

instance = model.instance_from_vector(vector=[0.1, 0.2, 0.3])

print("Model Instance: \n")
print(instance)

print("Instance Parameters \n")
print("x = ", instance.centre)
print("intensity = ", instance.intensity)
print("sigma = ", instance.sigma)

# %%
"""
Defining a model using multiple model components is straight forward in **PyAutoFit**, using a _CollectionPriorModel_
object.
"""

# %%
model = af.CollectionPriorModel(
    gaussian=htf.profiles.Gaussian, exponential=htf.profiles.Exponential
)

# %%
"""
A `CollectionPriorModel` behaves like a `PriorModel` but contains a collection of model components. For example, it can
create a model instance by mapping a list of parameters, which in this case is 6 (3 for the `Gaussian` `[centre,
intensity, sigma]` and 3 for the `Exponential` `[centre, intensity, rate]`).
"""

# %%
instance = model.instance_from_vector(vector=[0.1, 0.2, 0.3, 0.4, 0.5, 0.01])

# %%
"""
This `instance` contains each of the model components we defined above, using the input argument name of the
_CollectionoPriorModel_ to define the attributes in the `instance`:
"""

# %%
print("Instance Parameters \n")
print("x (Gaussian) = ", instance.gaussian.centre)
print("intensity (Gaussian) = ", instance.gaussian.intensity)
print("sigma (Gaussian) = ", instance.gaussian.sigma)
print("x (Exponential) = ", instance.exponential.centre)
print("intensity (Exponential) = ", instance.exponential.intensity)
print("sigma (Exponential) = ", instance.exponential.rate)

# %%
"""
We can call the components of a `CollectionPriorModel` whatever we want, and the mapped `instance` will use those names.
"""

# %%
model_custom_names = af.CollectionPriorModel(
    james=htf.profiles.Gaussian, rich=htf.profiles.Exponential
)

instance = model_custom_names.instance_from_vector(
    vector=[0.1, 0.2, 0.3, 0.4, 0.5, 0.01]
)

print("Instance Parameters \n")
print("x (Gaussian) = ", instance.james.centre)
print("intensity (Gaussian) = ", instance.james.intensity)
print("sigma (Gaussian) = ", instance.james.sigma)
print("x (Exponential) = ", instance.rich.centre)
print("intensity (Exponential) = ", instance.rich.intensity)
print("sigma (Exponential) = ", instance.rich.rate)

# %%
"""
Now we can create a model composed of multiple components, lets fit it to a `Dataset`. To do this, we updated this 
tutorial`s `phase` package, spefically its `Analysis` class such that it creates `model_data` as a super position of 
all the model`s individual `Profile`'s. For example, in the model above, the `model_data` is the sum of the 
_Gaussian_`s  individual profile and `Exponential`'s individual profile.

Checkout `phase.py` and `analysis.py` now, for a description of how this has been implemented.
"""

# %%
"""
Import the `simulators` module and set up the `Dataset`. This uses a new `Dataset` that is generated as a sum of a 
_Gaussian_ and `Exponential` profile.
"""

# %%
from howtofit.simulators.chapter_1 import gaussian_x1_exponential_x1

dataset = htf.Dataset(
    data=gaussian_x1_exponential_x1.data, noise_map=gaussian_x1_exponential_x1.noise_map
)

# %%
"""
We again need to create a `mask` for our data which we pass to the `phase.run` method. In this example, our `mask` is
only `False` entries meaning that every datapoint is used in the fit.
"""

# %%
mask = np.full(fill_value=False, shape=dataset.data.shape)

# %%
"""
Lets now perform the fit using our model which is composed of two _Profile`_s. You`ll note that the _Emcee_
dimensionality has increased from N=3 to N=6, given that we are now fitting two `Profile`'s each with 3 free parameters.
"""

# %%
phase = htf.Phase(
    phase_name="phase_t6_gaussian_x1_exponential_x1",
    profiles=af.CollectionPriorModel(
        gaussian=htf.profiles.Gaussian, exponential=htf.profiles.Exponential
    ),
    search=af.Emcee(),
)

print(
    "Emcee has begun running - checkout the autofit_workspace/howtofit/chapter_1_introduction/output/phase_t5_gaussian_x1_exponential_x1"
    " folder for live output of the results."
    "This Jupyter notebook cell with progress once Emcee has completed - this could take a few minutes!"
)

phase.run(dataset=dataset, mask=mask)

print("Emcee has finished run - you may now continue the notebook.")

# %%
"""
Quickly inspect the `Result`'s of the fit, which you may have noticed takes a bit longer to run than the fits performed
in previous tutorials. This is because the dimensionality of the model we are fitted increased from 3 to 6.

With the `CollectionPriorModel`, **PyAutoFit** provides all the tools needed to compose and fit any model imaginable!
Lets fit a model composed of two `Gaussian`. and and an `Exponential`, which will have a dimensionality of N=9.
"""

# %%
from howtofit.simulators.chapter_1 import gaussian_x2_exponential_x1

dataset = htf.Dataset(
    data=gaussian_x2_exponential_x1.data, noise_map=gaussian_x2_exponential_x1.noise_map
)

phase = htf.Phase(
    phase_name="phase_t6_gaussian_x2_exponential_x1",
    profiles=af.CollectionPriorModel(
        gaussian_0=htf.profiles.Gaussian,
        gaussian_1=htf.profiles.Gaussian,
        exponential=htf.profiles.Exponential,
    ),
    search=af.Emcee(),
)

print(
    "Emcee has begun running - checkout the autofit_workspace/howtofit/chapter_1_introduction/output/phase_t5_gaussian_x2_exponential_x1"
    " folder for live output of the results."
    "This Jupyter notebook cell with progress once Emcee has completed - this could take a few minutes!"
)

phase.run(dataset=dataset, mask=mask)

print("Emcee has finished run - you may now continue the notebook.")

# %%
"""
We can fully customize the model that we fit. Lets suppose we have a `Dataset` that consists of three `Gaussian` 
profiles, but we also know the following information about the dataset:

- All 3 `Gaussian`.s are centrally aligned.
- The `sigma` of one `Gaussian` is equal to 1.0.

We can edit our `CollectionPriorModel` to meet these constraints accordingly:
"""

# %%
model = af.CollectionPriorModel(
    gaussian_0=htf.profiles.Gaussian,
    gaussian_1=htf.profiles.Gaussian,
    gaussian_2=htf.profiles.Gaussian,
)

# %%
"""
This aligns the `centre`'s of the 3 `Gaussian`., reducing the dimensionality of the model from N=9 to N=7
"""

# %%
model.gaussian_0.centre = model.gaussian_1.centre
model.gaussian_1.centre = model.gaussian_2.centre

# %%
"""
This fixes the `sigma` value of one `Gaussian` to 1.0, further reducing the dimensionality from N=7 to N=6.
"""

# %%
model.gaussian_0.sigma = 1.0

# %%
"""
We can now fit this model using a `Phase` as per usual.
"""

# %%
from howtofit.simulators.chapter_1 import gaussian_x3

dataset = htf.Dataset(data=gaussian_x3.data, noise_map=gaussian_x3.noise_map)

phase = htf.Phase(phase_name="phase_t6_gaussian_x3", profiles=model, search=af.Emcee())

print(
    "Emcee has begun running - checkout the autofit_workspace/howtofit/chapter_1_introduction/output/phase_t5_gaussian_x3"
    " folder for live output of the results."
    "This Jupyter notebook cell with progress once Emcee has completed - this could take a few minutes!"
)

phase.run(dataset=dataset, mask=mask)

print("Emcee has finished run - you may now continue the notebook.")

# %%
"""
And with that, we are complete. In this tutorial, we learned how to compose complex models in **PyAutoFit** and adjust 
our `phase.py` and `analyis.py` modules to fit them. To end, you should think again in more detail about your model
fitting problem:

Are there many different model components you may wish to define and fit?

Is your data the super position of many different model components, like the profiles in this tutorial?

In this tutorial, all components of our model did the same thing, represent a 1D profile. In your model, you may
have model components that represent different parts of your model, which need to be combined in more complicated ways
in order to create your model-fit. In such circumstances, the `fit` method in your `Analysis` class may be 
significantly more complex than the example shown in this tutorial. Nevertheless, you now have all the tools you need 
to define, compose and fit very complex models, there isn't much left for you to learn on your journey through **PyAutoFit**!
"""



import os

def check_workspace_path():
    if "WORKSPACE" not in os.environ:
        raise OSError(
            "\n\n#############################################\n"
            "### WORKSPACE ENVIRONMENT VARIABLE ERRROR ###\n"
            "#############################################\n\n"
            "PyAutoFit is Unable to locate the WORKSPACE environment variable. This variable must be set for "
            "PyAutoFit to know the location of the configuration files and results output folder.\n\n"
            ""
            "To set the WORKSPACE environment variable, run the 'setup_environment.py' script in the "
            "autofit_workspace. If you have not yet downloaded the autofit_workspace do this using the link below, for "
            "example if you cloned / forked the PyAutoFit repository. The GitHub repo for the autofit_workspace is:\n\n"
            ""
            "https://github.com/Jammy2211/autofit_workspace\n\n"
            ""
            "For more information on the WORKSPACE environment variable and instructions for settings it manually, see "
            "the link below:\n\n"
            ""
            "https://pyautofit.readthedocs.io/en/latest/general/installation.html#environment-variables\n\n"
        )

    from autoconf import conf

    try:

        conf.get_matplotlib_backend()
        return

    except Exception:

        raise OSError(
            "\n\n#############################################\n"
            "### WORKSPACE ENVIRONMENT VARIABLE ERRROR ###\n"
            "#############################################\n\n"
            "PyAutoFit located the following WORKSPACE environment variable:\n\n"
            ""
            f"{os.environ['WORKSPACE']}\n\n"
            ""
            "This path does not point to an autofit_workspace on your computer. Please double check the path and "
            "update the WORKSPACE path to point to a valid autofit_workspace.\n\n"

            "To reset the WORKSPACE environment variable, run the 'setup_environment.py' script in the "
            "autofit_workspace. If you have not yet downloaded the autofit_workspace do this using the link below, "
            "for example if you cloned / forked the PyAutoFit repository. The GitHub repo for the autofit_workspace is:\n\n"
            ""
            "https://github.com/Jammy2211/autofit_workspace\n\n"
            ""
            "For more information on the WORKSPACE environment variable and instructions for settings it manually, see "
            "the link below:\n\n"
            ""
            "https://pyautofit.readthedocs.io/en/latest/general/installation.html#environment-variables\n\n"
        )

from autoconf.conf import *


from typing import Tuple, Dict, NamedTuple

import numpy as np

from autofit.mapper.variable import Variable


class JacobianValue(NamedTuple):
    log_value: Dict[str, np.ndarray]
    deterministic_values: Dict[Tuple[str, str], np.ndarray]


def numerical_jacobian(
        factor,
        args: Tuple[Variable, ...],
        kwargs: Dict[Variable, np.array],
        _eps: float = 1e-6,
        _calc_deterministic: bool = True,
) -> JacobianValue:
    """Calculates the numerical Jacobian of the passed factor

    the arguments passed correspond to the variable that we want
    to take the derivatives for

    the values must be passed as keywords

    _eps specifies the numerical accuracy of the numerical derivative

    returns a jac = JacobianValue namedtuple

    jac.log_value stores the Jacobian of the factor output as a dictionary
    where the keys are the names of the variables

    jac.determinisic_values store the Jacobian of the deterministic variables
    where the keys are a Tuple[str, str] pair of the variable and deterministic
    variables

    Example
    -------
    >>> import numpy as np
    >>> y_ = Variable('y')
    >>> A = np.arange(4).reshape(2, 2)
    >>> dot = factor(lambda x: A.dot(x))(x_) == y_
    >>> dot.jacobian('x', x=[1, 2])
    JacobianValue(
        log_value={'x': array([[0.], [0.]])},
        deterministic_values={('x', 'y'): array([[0., 2.], [1., 3.]])})
    """
    # copy the input array
    p0 = {v: np.array(x, dtype=float) for v, x in kwargs.items()}
    f0 = factor(p0)
    log_f0 = f0.log_value
    det_vars0 = f0.deterministic_values

    jac_f = {
        v: np.empty(np.shape(kwargs[v]) + np.shape(log_f0))
        for v in args}
    if _calc_deterministic:
        jac_det = {
            (det, v): np.empty(
                np.shape(val) + np.shape(kwargs[v]))
            for v in args
            for det, val in det_vars0.items()}
        det_slices = {
            v: (slice(None),) * np.ndim(a) for v, a in kwargs.items()}
    else:
        jac_det = {}

    for v in args:
        x0 = p0[v]
        if x0.shape:
            inds = tuple(a.ravel() for a in np.indices(x0.shape))
            for ind in zip(*inds):
                x0[ind] += _eps
                p0[v] = x0
                f = factor(p0)
                x0[ind] -= _eps

                jac_f[v][ind] = (f.log_value - log_f0) / _eps
                if _calc_deterministic:
                    det_vars = f.deterministic_values
                    for det, val in det_vars.items():
                        jac_det[det, v][det_slices[v] + ind] = \
                            (val - det_vars0[det]) / _eps
        else:
            p0[v] += _eps
            f = factor(p0)
            p0[v] -= _eps

            jac_f[v] = (f.log_value - log_f0) / _eps
            if _calc_deterministic:
                det_vars = f.deterministic_values
                for det, val in det_vars.items():
                    jac_det[det, v] = (val - det_vars0[det]) / _eps

    return JacobianValue(jac_f, jac_det)



import logging
import pickle
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Dict
import dill

from autoconf import conf
from autofit.mapper.model_mapper import ModelMapper
from autofit.non_linear.paths import convert_paths
from autofit.mapper.prior.promise import PromiseResult
from autofit.non_linear import grid_search
from autofit.non_linear.paths import Paths

logger = logging.getLogger(__name__)


class AbstractPhase:
    @convert_paths
    def __init__(
            self,
            paths: Paths,
            *,
            search,
            model=None,
    ):
        """
        A phase in an lens pipeline. Uses the set non_linear search to try to
        fit_normal models and image passed to it.

        Parameters
        ----------
        search: class
            The class of a non_linear search
        """

        self.search = search
        self.model = model or ModelMapper()

        self.pipeline_name = None
        self.pipeline_tag = None

    @property
    def paths(self):
        return self.search.paths

    def save_model_info(self):
        """Save the model.info file, which summarizes every parameter and prior."""
        with open(self.paths.file_model_info, "w+") as f:
            f.write(self.model.info)

    @property
    def _default_metadata(self) -> Dict[str, str]:
        """
        A dictionary of metadata describing this phase, including the pipeline
        that it's embedded in.
        """
        return {
            "phase": self.paths.name,
            "phase_tag": self.paths.tag,
            "pipeline": self.pipeline_name,
            "pipeline_tag": self.pipeline_tag,
        }

    def make_metadata_text(self, dataset_name):
        return "\n".join(
            f"{key}={value or ''}"
            for key, value
            in {
                **self._default_metadata,
                "dataset_name": dataset_name
            }.items()
        )

    def save_metadata(self, dataset):
        """
        Save metadata associated with the phase, such as the name of the pipeline, the
        name of the phase and the name of the dataset being fit
        """
        with open("{}/metadata".format(self.paths.make_path()), "w+") as f:
            f.write(
                self.make_metadata_text(
                    dataset.name
                )
            )

    def save_dataset(self, dataset):
        """
        Save the dataset associated with the phase
        """
        with open(f"{self.paths.pickle_path}/dataset.pickle", "wb") as f:
            pickle.dump(dataset, f)

    def save_mask(self, mask):
        """
        Save the mask associated with the phase
        """
        with open(f"{self.paths.pickle_path}/mask.pickle", "wb") as f:
            dill.dump(mask, f)

    def save_settings(self, settings):
        with open(
                f"{self.paths.pickle_path}/settings.pickle",
                "wb+"
        ) as f:
            pickle.dump(
                settings, f
            )

    def save_phase_attributes(self, phase_attributes):
        with open(
                f"{self.paths.pickle_path}/phase_attributes.pickle",
                "wb+"
        ) as f:
            pickle.dump(
                phase_attributes, f
            )

    def __str__(self):
        return self.search.paths.name

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.search.paths.name}>"

    @property
    def result(self) -> PromiseResult:
        """
        A PromiseResult allows promises to be defined, which express the equality
        between posteriors or best fits from this phase and priors or constants
        in some subsequent phase.
        """
        return PromiseResult(self)

    def run_analysis(self, analysis, info=None, pickle_files=None):
        return self.search.fit(model=self.model, analysis=analysis, info=info, pickle_files=pickle_files)

    def make_phase_attributes(self, analysis):
        raise NotImplementedError()

    def make_result(self, result, analysis):
        raise NotImplementedError()

    @property
    def phase_name(self):
        return self.paths.name


class Dataset(ABC):
    """
    Comprises the data that is fit by the pipeline. May also contain meta data, noise, PSF, etc.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """
        The name of this data for use in querying
        """

    @classmethod
    def load(cls, filename) -> "Dataset":
        """
        Load the dataset at the specified filename

        Parameters
        ----------
        filename
            The filename containing the dataset

        Returns
        -------
        The dataset
        """
        with open(filename, "rb") as f:
            return pickle.load(f)


class Phase(AbstractPhase):
    @convert_paths
    def __init__(
            self,
            paths,
            *,
            analysis_class,
            search,
            model=None,
    ):
        super().__init__(paths=paths, search=search, model=model)
        self.analysis_class = analysis_class

    def make_result(self, result, analysis):
        return result

    def make_analysis(self, dataset):
        return self.analysis_class(dataset)

    def run(self, dataset: Dataset, results=None, info=None, pickle_files=None):
        """
        Run this phase.

        Parameters
        ----------
        results: autofit.tools.pipeline.ResultsCollection
            An object describing the results of the last phase or None if no phase has been executed
        dataset: scaled_array.ScaledSquarePixelArray
            An masked_imaging that has been masked

        Returns
        -------
        result: AbstractPhase.Result
            A result object comprising the best fit model and other hyper_galaxies.
        """
        self.save_metadata(dataset=dataset)
        self.save_dataset(dataset=dataset)

        self.model = self.model.populate(results)

        analysis = self.make_analysis(dataset=dataset)

        result = self.run_analysis(analysis=analysis, info=info, pickle_files=pickle_files)

        return self.make_result(result=result, analysis=None)


def as_grid_search(phase_class, parallel=False):
    """
    Create a grid search phase class from a regular phase class. Instead of the phase
    being optimised by a single non-linear optimiser, a new optimiser is created for
    each square in a grid.

    Parameters
    ----------
    phase_class
        The original phase class
    parallel: bool
        Indicates whether non linear searches in the grid should be performed on
        parallel processes.

    Returns
    -------
    grid_search_phase_class: GridSearchExtension
        A class that inherits from the original class, replacing the optimiser with a
        grid search optimiser.

    """

    class GridSearchExtension(phase_class):
        @convert_paths
        def __init__(
                self,
                paths,
                *,
                search,
                number_of_steps=4,
                **kwargs,
        ):

            super().__init__(paths, search=search, **kwargs)

            self.search = grid_search.GridSearch(
                paths=self.paths,
                number_of_steps=number_of_steps,
                search=search,
                parallel=parallel,
            )

        def save_grid_search_result(self, grid_search_result):
            with open(
                    f"{self.paths.pickle_path}/grid_search_result.pickle",
                    "wb+"
            ) as f:
                pickle.dump(
                    grid_search_result, f
                )

        # noinspection PyMethodMayBeStatic,PyUnusedLocal
        def make_result(self, result, analysis):
            self.save_grid_search_result(grid_search_result=result)
            open(self.paths.has_completed_path, "w+").close()

            return self.Result(
                samples=result.samples,
                previous_model=result.model,
                analysis=analysis,
                search=self.search,
            )

        def run_analysis(self, analysis, **kwargs):

            self.search.search.paths = self.paths
            self.search.paths = self.paths

            return self.search.fit(model=self.model, analysis=analysis, grid_priors=self.grid_priors)

        @property
        def grid_priors(self):
            raise NotImplementedError(
                "The grid priors property must be implemented to provide a list of "
                "priors to be grid searched"
            )

    return GridSearchExtension


class AbstractSettingsPhase:

    def __init__(self, log_likelihood_cap=None):

        self.log_likelihood_cap = log_likelihood_cap

    @property
    def log_likelihood_cap_tag(self):
        """Generate a bin up tag, to customize phase names based on the resolutioon the image is binned up by for faster \
        run times.

        This changes the phase settings folder is tagged as follows:

        bin_up_factor = 1 -> settings
        bin_up_factor = 2 -> settings_bin_up_factor_2
        bin_up_factor = 2 -> settings_bin_up_factor_2
        """
        if self.log_likelihood_cap is None:
            return ""
        return f"__{conf.instance.settings_tag.get('phase', 'log_likelihood_cap')}" \
                + "_{0:.1f}".format(self.log_likelihood_cap)


import os
import pickle

import dill

from autofit.non_linear import abstract_search
from autofit.non_linear.samples import PDFSamples


class PhaseOutput:
    """
    @DynamicAttrs
    """

    def __init__(self, directory: str):
        """
        Represents the output of a single phase. Comprises a metadata file and other dataset files.

        Parameters
        ----------
        directory
            The directory of the phase
        """
        self.directory = directory
        self.__search = None
        self.__model = None
        self.file_path = os.path.join(directory, "metadata")
        with open(self.file_path) as f:
            self.text = f.read()
            pairs = [
                line.split("=")
                for line
                in self.text.split("\n")
                if "=" in line
            ]
            self.__dict__.update({pair[0]: pair[1] for pair in pairs})

    @property
    def pickle_path(self):
        return f"{self.directory}/pickles"

    @property
    def model_results(self) -> str:
        """
        Reads the model.results file
        """
        with open(os.path.join(self.directory, "model.results")) as f:
            return f.read()

    @property
    def mask(self):
        """
        A pickled mask object
        """
        with open(
                os.path.join(self.pickle_path, "mask.pickle"), "rb"
        ) as f:
            return dill.load(f)

    def __getattr__(self, item):
        """
        Attempt to load a pickle by the same name from the phase output directory.

        dataset.pickle, meta_dataset.pickle etc.
        """
        try:
            with open(
                    os.path.join(self.pickle_path, f"{item}.pickle"), "rb"
            ) as f:
                return pickle.load(f)
        except FileNotFoundError:
        #    print(f"No {item} associated with {self.directory}")
            pass

    @property
    def header(self) -> str:
        """
        A header created by joining the pipeline, phase and dataset names
        """
        return "/".join((self.pipeline, self.phase, self.dataset_name))

    @property
    def search(self) -> abstract_search.NonLinearSearch:
        """
        The search object that was used in this phase
        """
        if self.__search is None:
            try:
                with open(os.path.join(self.pickle_path, "search.pickle"), "r+b") as f:
                    self.__search = pickle.loads(f.read())
            except FileNotFoundError:
                pass
        return self.__search

    @property
    def model(self):
        """
        The model that was used in this phase
        """
        if self.__model is None:
            with open(os.path.join(self.pickle_path, "model.pickle"), "r+b") as f:
                self.__model = pickle.loads(f.read())
        return self.__model

    def __str__(self):
        return self.text

    def __repr__(self):
        return "<PhaseOutput {}>".format(self)



import numpy as np
from scipy import special

from autofit.graphical.messages.abstract import AbstractMessage
from autofit.graphical.utils import invpsilog


class GammaMessage(AbstractMessage):
    @property
    def log_partition(self):
        alpha, beta = GammaMessage.invert_natural_parameters(
            self.natural_parameters
        )
        return special.gammaln(alpha) - alpha * np.log(beta)

    log_base_measure = 0.
    _support = ((0, np.inf),)
    _parameter_support = ((0, np.inf), (0, np.inf))

    def __init__(
            self,
            alpha=1.,
            beta=1.,
            log_norm=0.
    ):
        self.alpha = alpha
        self.beta = beta
        super().__init__(
            parameters=[
                alpha, beta
            ],
            log_norm=log_norm
        )

    @property
    def natural_parameters(self):
        return self.calc_natural_parameters(
            self.alpha,
            self.beta
        )

    @staticmethod
    def calc_natural_parameters(alpha, beta):
        return np.array([alpha - 1, - beta])

    @staticmethod
    def invert_natural_parameters(natural_parameters):
        eta1, eta2 = natural_parameters
        return eta1 + 1, -eta2

    @staticmethod
    def to_canonical_form(x):
        return np.array([np.log(x), x])

    @classmethod
    def invert_sufficient_statistics(cls, suff_stats):
        logX, X = suff_stats
        alpha = invpsilog(logX - np.log(X))
        beta = alpha / X
        return cls.calc_natural_parameters(alpha, beta)

    @property
    def mean(self):
        return self.alpha / self.beta

    @property
    def variance(self):
        return self.alpha / self.beta ** 2

    def sample(self, n_samples):
        a1, b1 = self.parameters
        return np.random.gamma(a1, scale=1 / b1, size=(n_samples,) + self.shape)

    @classmethod
    def from_mode(cls, mode, covariance):
        m, V = cls._get_mean_variance(mode, covariance)

        alpha = 1 + m ** 2 * V  # match variance
        beta = alpha / m  # match mean
        return cls(alpha, beta)


from autofit import exc

from autofit.non_linear.log import logger

import configparser


class Initializer:
    def __init__(self, lower_limit, upper_limit):
        """
        The Initializer creates the initial set of samples in non-linear parameter space that can be passed into a
        non-linear search to define where to begin sampling.

        Although most non-linear searches have in-built functionality to do this, some do not cope well with parameter
        resamples that are raised as FitException's. Thus, PyAutoFit uses its own initializer to bypass these problems.

        Parameters
        ----------
        """
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit

    @classmethod
    def from_config(cls, config):
        """Load the Initializer from a non_linear config file."""

        try:

            initializer = config("initialize", "method", str)

        except configparser.NoSectionError:

            return None

        if initializer in "prior":

            return InitializerPrior()

        elif initializer in "ball":

            ball_lower_limit = config("initialize", "ball_lower_limit", float)
            ball_upper_limit = config("initialize", "ball_upper_limit", float)

            return InitializerBall(
                lower_limit=ball_lower_limit, upper_limit=ball_upper_limit
            )

    def initial_samples_from_model(self, total_points, model, fitness_function):
        """Generate the initial points of the non-linear search, by randomly drawing unit values from a uniform
        distribution between the ball_lower_limit and ball_upper_limit values.

        Parameters
        ----------
        total_points : int
            The number of points in non-linear paramemter space which initial points are created for.
        model : ModelMapper
            An object that represents possible instances of some model with a given dimensionality which is the number
            of free dimensions of the model.
        """

        logger.info("Generating initial samples of model, which are subject to prior limits and other constraints.")

        initial_unit_parameters = []
        initial_parameters = []
        initial_figures_of_merit = []

        point_index = 0

        while point_index < total_points:

            unit_parameters = model.random_unit_vector_within_limits(
                lower_limit=self.lower_limit, upper_limit=self.upper_limit
            )
            parameters = model.vector_from_unit_vector(unit_vector=unit_parameters)

            try:
                figure_of_merit = fitness_function.figure_of_merit_from_parameters(
                    parameters=parameters
                )
                initial_unit_parameters.append(unit_parameters)
                initial_parameters.append(parameters)
                initial_figures_of_merit.append(figure_of_merit)
                point_index += 1
            except exc.FitException:
                pass

        return initial_unit_parameters, initial_parameters, initial_figures_of_merit


class InitializerPrior(Initializer):
    def __init__(self):
        """
        The Initializer creates the initial set of samples in non-linear parameter space that can be passed into a
        non-linear search to define where to begin sampling.

        Although most non-linear searches have in-built functionality to do this, some do not cope well with parameter
        resamples that are raised as FitException's. Thus, PyAutoFit uses its own initializer to bypass these problems.

        The InitializerPrior class generates from the priors, by drawing all values as unit values between 0.0 and 1.0
        and mapping them to physical values via the prior.
        """

        super().__init__(lower_limit=0.0, upper_limit=1.0)


class InitializerBall(Initializer):
    def __init__(self, lower_limit, upper_limit):
        """
        The Initializer creates the initial set of samples in non-linear parameter space that can be passed into a
        non-linear search to define where to begin sampling.

        Although most non-linear searches have in-built functionality to do this, some do not cope well with parameter
        resamples that are raised as FitException's. Thus, PyAutoFit uses its own initializer to bypass these problems.

        The InitializerBall class generates the samples in a small compact volume or 'ball' in parameter space, which is
        the recommended initialization strategy for the MCMC non-linear search Emcee.

        Parameters
        ----------
        lower_limit : float
            The lower limit of the uniform distribution unit values are drawn from when initializing walkers in a small
            compact ball.
        upper_limit : float
            The upper limit of the uniform distribution unit values are drawn from when initializing walkers in a small
            compact ball.
        """

        super().__init__(lower_limit=lower_limit, upper_limit=upper_limit)



import numpy as np

"""
This is our dataset, which in tutorial 1 stores a 1D `Gaussian` with a noise-map.

Our data is a 1D numpy array of values corresponding to the observed counts of the Gaussian.
The noise-map corresponds to the expected noise in every data point.

In PyAutoFit we use the term `dataset` to represent the collection of data, noise-map, etc. that are fitted. In this
chapter our `Dataset` is a 1D line, for your model-fitting problem your `Dataset` is probably more complex than this.

In these tutorials we are calling our `Dataset` class `Dataset`, as opposed to something more specific to the data
(e.g. `Line`). We advise that the `Dataset` class is give a more specific name (e.g., if your `Dataset` is imaging data,
the class may be called `Imaging`. If your `Dataset` is a spectrum, you may use `Spectrum`).

However, to make the tutorials clear, we've stuck with the name `Dataset` for clarity. The example project at the
end of the chapter will adopt a more specific `Dataset` naming convention.
"""


class Dataset:
    def __init__(self, data, noise_map):
        """A class containing the data and noise-map of a 1D line `Dataset`.

        Parameters
        ----------
        data : np.ndarray
            The array of the data, in arbitrary units.
        noise_map : np.ndarray
            An array describing the RMS standard deviation error in each data pixel, in arbitrary units.
        """
        self.data = data
        self.noise_map = noise_map

    """The `xvalues` defines the x coordinates of the 1D Gaussian, which we assume run from 0 the length of the data."""

    @property
    def xvalues(self):
        return np.arange(self.data.shape[0])



from howtofit.chapter_1_introduction.tutorial_4_source_code.src.plot import (
    line_plots,
)

"""
These functions are simple matplotlib calls that plot components of our Line class, specifically its data and
noise-map.

Storing simple functions like this for plotting components of our `Dataset` will prove beneficial in later tutorials,
when it comes to inspecting the results of a model-fit after they have been completed.
"""


def data(dataset):
    """Plot the data values of a Line `Dataset`.

    Parameters
    -----------
    Line : `Dataset`.Line
        The observed Line `Dataset` whose data is plotted.
    """
    line_plots.line(xvalues=dataset.xvalues, line=dataset.data, ylabel="Data Values")


def noise_map(dataset):
    """Plot the noise-map of a Line `Dataset`.

    Parameters
    -----------
    Line : `Dataset`.Line
        The observed Line `Dataset` whose data is plotted.
    """
    line_plots.line(xvalues=dataset.xvalues, line=dataset.noise_map, ylabel="Noise-Map")



import logging

from autoconf import conf
from autofit.text import formatter as frm, samples_text


def results_to_file(samples, filename, during_analysis):
    """Output the full model.results file, which include the most-likely model, most-probable model at 1 and 3
    sigma confidence and information on the maximum log likelihood.

    Parameters
    ----------
    during_analysis : bool
        Whether the model.results are being written during the analysis or after the non-linear search has finished.
    """

    results = []

    if hasattr(samples, "log_evidence"):
        if samples.log_evidence is not None:

            value = "{:.8f}".format(samples.log_evidence)
            results += [
                frm.add_whitespace(str0="Bayesian Evidence ", str1=value, whitespace=90)
            ]
            results += ["\n"]

    value = "{:.8f}".format(max(samples.log_likelihoods))
    results += [
        frm.add_whitespace(str0="Maximum Likelihood ", str1=value, whitespace=90)
    ]
    results += ["\n\n"]

    results += ["Maximum Log Likelihood Model:\n\n"]

    formatter = frm.TextFormatter()

    for i, prior_path in enumerate(samples.model.unique_prior_paths):
        formatter.add(
            (prior_path, format_str().format(samples.max_log_likelihood_vector[i]))
        )
    results += [formatter.text + "\n"]

    if hasattr(samples, "pdf_converged"):

        if samples.pdf_converged:

            results += samples_text.summary(samples=samples, sigma=3.0, indent=4, line_length=90)
            results += ["\n"]
            results += samples_text.summary(samples=samples, sigma=1.0, indent=4, line_length=90)

        else:

            results += [
                "\n WARNING: The samples have not converged enough to compute a PDF and model errors. \n "
                "The model below over estimates errors. \n\n"
            ]
            results += samples_text.summary(samples=samples, sigma=1.0, indent=4, line_length=90)

        results += ["\n\ninstances\n"]

    formatter = frm.TextFormatter()

    for t in samples.model.path_float_tuples:
        formatter.add(t)

    results += ["\n" + formatter.text]

    frm.output_list_of_strings_to_file(file=filename, list_of_strings=results)


def search_summary_from_samples(samples) -> [str]:

    line = [f"Total Samples = {samples.total_samples}\n"]
    if hasattr(samples, "total_accepted_samples"):
        line.append(f"Total Accepted Samples = {samples.total_accepted_samples}\n")
        line.append(f"Acceptance Ratio = {samples.acceptance_ratio}\n")
    if samples.time is not None:
        line.append(f"Time To Run = {samples.time}\n")
    return line


def search_summary_to_file(samples, filename):

    summary = search_summary_from_samples(samples=samples)

    frm.output_list_of_strings_to_file(file=filename, list_of_strings=summary)


def format_str() -> str:
    """The format string for the model.results file, describing to how many decimal points every parameter
    estimate is output in the model.results file.
    """
    decimal_places = conf.instance.general.get(
        "output", "model_results_decimal_places", int
    )
    return f"{{:.{decimal_places}f}}"



import os

import pytest

import autofit as af
from test_autofit.mock import MockPhaseOutput


@pytest.fixture(
    name="path_aggregator"
)
def make_path_aggregator(
        aggregator_directory
):
    aggregator = af.Aggregator(
        aggregator_directory
    )
    yield aggregator
    aggregator.remove_unzipped()


@pytest.fixture(
    name="aggregator_directory"
)
def make_aggregator_directory():
    directory = os.path.dirname(
        os.path.realpath(__file__)
    )
    return f"{directory}/../tools/files/aggregator"


@pytest.fixture(name="aggregator")
def make_aggregator():
    aggregator = af.Aggregator("")
    aggregator.phases = [
        MockPhaseOutput(
            "directory/number/one",
            "pipeline1",
            "phase1",
            "dataset1"
        ),
        MockPhaseOutput(
            "directory/number/two",
            "pipeline1",
            "phase2",
            "dataset1"
        ),
        MockPhaseOutput(
            "directory/letter/a",
            "pipeline2",
            "phase2",
            "dataset2"
        ),
    ]
    return aggregator



import autofit as af

"""The `result.py` module is unchanged from the previous tutorial."""


class Result(af.Result):
    def __init__(self, samples, previous_model, search, analysis):
        """
        The results of a `NonLinearSearch` performed by a phase.

        Parameters
        ----------
        samples : af.Samples
            A class containing the samples of the `NonLinearSearch`, including methods to get the maximum log
            likelihood model, errors, etc.
        analysis : Analysis
            The Analysis class used by this model-fit to fit the model to the data.
        """
        super().__init__(samples=samples, previous_model=previous_model, search=search)
        self.analysis = analysis

    @property
    def max_log_likelihood_model_data(self):
        return self.analysis.model_data_from_instance(
            instance=self.samples.max_log_likelihood_instance
        )

    @property
    def max_log_likelihood_fit(self):
        return self.analysis.fit_from_model_data(
            model_data=self.max_log_likelihood_model_data
        )



from autofit.graphical.factor_graphs.abstract import AbstractNode
from autofit.graphical.factor_graphs.factor import FactorValue, Factor, AbstractFactor
from autofit.graphical.factor_graphs.graph import DeterministicFactorNode, FactorGraph
from autofit.graphical.factor_graphs.numerical import numerical_jacobian
from autofit.mapper.variable import Variable, Plate



# %%
"""
Tutorial 5: Visualization & Masking
===================================

In the previous tutorial, we wrote source code which used **PyAutoFit** to fit a 1D `Gaussian` model to a `Dataset`.
In this tutorial, we'll extend this source code`s phase package to perform a number of additional tasks:

 - Masking: The phase is passed a mask such that regions of the `Dataset` are omitted by the `log_likelihood_function`.

 - Visualization: Images showing the model fit are output on-the-fly during the `NonLinearSearch`.

These new features have lead to an additional module in the `phase` package not present in tutorial 4, called
`visualizer.py`. Before looking at this module, lets perform a fit to see the changed behaviour of **PyAutoFit**.
"""

# %%
#%matplotlib inline

from autoconf import conf
import autofit as af
from howtofit.chapter_1_introduction.tutorial_5_visualization_masking import (
    src as htf,
)

import numpy as np
import os

workspace_path = os.environ["WORKSPACE"]
print("Workspace Path: ", workspace_path)

# %%
"""
Setup the configs as we did in the previous tutorial, as well as the output folder for our `NonLinearSearch`.
"""

# %%
conf.instance = conf.Config(
    config_path=f"{workspace_path}/config",
    output_path=f"{workspace_path}/output/chapter_1",
)

# %%
"""
Import the `simulators` module and set up the `Dataset`.
"""

# %%
from howtofit.simulators.chapter_1 import gaussian_x1

dataset = htf.Dataset(data=gaussian_x1.data, noise_map=gaussian_x1.noise_map)

# %%
"""
Before fitting data, we may want to mask it, removing regions of the data we know are defective or where there is no
signal.

To facilitate this we have added a new class to the module `dataset.py`. This takes our `Dataset` and a mask and 
combines the two to create a `MaskedDataset`. The `fit.py` module has also been updated to use a mask during the fit. 
Check them both out now to see how the mask is used! 

As mentioned in tutorial 4, if your model-fitting problem involves fitting masked data, you should be able to copy and 
paste the `fit.py` module for your own project.

Masking occurs within the phase package of **PyAutoFit**, which we'll inspect at the end of the tutorial. However,
for a phase to run it now requires that a `mask` is passed to it. For this tutorial, lets create a which removes the
last 30 data-points in our `data`.

(In our convention, a `mask` value of `True` means it IS masked and thus removed from the fit).
"""

# %%
mask = np.full(fill_value=False, shape=dataset.data.shape)
mask[-30:] = True

# %%
"""
Lets now perform the fit from tutorial 4, but with a `MaskedDataset` and visualization.
"""

# %%
phase = htf.Phase(
    phase_name="phase_t5", gaussian=af.PriorModel(htf.Gaussian), search=af.Emcee()
)

print(
    "Emcee has begun running - checkout the autofit_workspace/howtofit/chapter_1_introduction/output/phase_t4"
    " folder for live output of the results."
    "This Jupyter notebook cell with progress once Emcee has completed - this could take a few minutes!"
)

# %%
"""
Note that we are passing our `mask` to the `phase.run` function, which we did not in previous tutorials.
"""

# %%
phase.run(dataset=dataset, mask=mask)

print("Emcee has finished run - you may now continue the notebook.")

# %%
"""
Lets check that this phase performs visualization. Navigate to the folder `image` in the directory above. You should 
see a set of folders containing the visualization of the `Dataset` and the `Fit`. As promised, our phase is
now taking care of the visualization of our model.

Visualization happens `on-the-fly`, such that during `Emcee` these images are output using the current maximum 
likelihood model `Emcee` has found. For models more complex than our 1D `Gaussian` this is useful, as it means we can 
check that `Emcee` has found reasonable solutions during a run and can thus cancel it early if it has ended up with an
incorrect solution.

How often does **PyAutoFit** output new images? This is set by `visualize_every_update` in the config file
`config/visualize/general.ini`.

Finally, now inspect the `phase.py`, `analysis.py` and `visualizer.py` scripts in the source code. These describe how 
the `MaskedData` is set up and how visualization is performed.

And with that, we have completed this (fairly short) tutorial. There are two things worth ending on:

 1) In tutorial 4, we introduced the `plot` package that had functions specific to plotting attributes of
 a data-set and fit. This project structure has again helped us, by making it straight forward to perform plotting with 
 the _Visualizer-. For your model-fitting project you should aim to adhere to performing all plots in a `plot` 
 module, as more benefits will become clear in chapter 2!
    
 2) For our very simple 1D case, we used a 1D NumPy array to represent a `mask`. For projects with more complicated
 datasets, one may require more complicated masks, warranting a `mask` package, `mask.py` module and `Mask2D` class. In 
 tutorial 9 we will show an example of this.
"""



import matplotlib.pyplot as plt

"""This function plots a line of 1D data, and was already introduce in tutorial 2."""


def line(xvalues, line, ylabel=None):
    """Plot a 1D line of data on a plot of x versus y, where the x-axis is the x coordinate of the line and the y-axis
    is the intensity of the line at that coordinate.

    Parameters
    ----------
    xvalues : ndarray
        The x-coordinates the profile is defined on.
    line : ndarray
        The intensity values of the profile which are plotted.
    ylabel : str
        The y-label of the plot.
    """

    plt.plot(xvalues, line)
    plt.xlabel("x")
    plt.ylabel(ylabel)
    plt.show()
    plt.clf()



# %%
"""
Tutorial 7: Phase Customization
===============================

In this tutorial, we`re going to add input parameters to a `Phase` object that customizes the analysis. we'll use the
specific example of two input parameters that trim our `Dataset` from the left and right before fitting it. This
example is somewhat trivial (we could achieve almost the same effect with masking), but it will serve to illustrate
`Phase` customization.

When we customize a `Phase`, we'll use `SettingsPhase` to perform `tagging`. Here, we `tag` the output
path of the `Phase`'s `Result``., such that every time a `Phase` is run with a different customization a new set of
unique `Result`'s are stored for those `PhaseSetting``.. For a given `Dataset` we are thus able to fit it multiple
times using different settings to compare the results.

These new features have led to an additional module in the `phase` package called `settings.py`, as well as extensions
to the `dataset.py` module. Before looking at these modules, lets first perform a series of `Emcee` fits to see how
they change the behaviour of **PyAutoFit**.
"""

# %%
#%matplotlib inline

from autoconf import conf
import autofit as af
import numpy as np

from howtofit.chapter_1_introduction.tutorial_7_phase_customization import (
    src as htf,
)

import os

workspace_path = os.environ["WORKSPACE"]
print("Workspace Path: ", workspace_path)

# %%
"""
Setup the configs as we did in the previous tutorial, as well as the output folder for our `NonLinearSearch`.
"""

# %%
conf.instance = conf.Config(
    config_path=f"{workspace_path}/config",
    output_path=f"{workspace_path}/output/chapter_1",
)

# %%
"""
We`re now going to perform multiple fits, where each fit trims the `Dataset` that is fitted.

To do this, we'll set up `Phase`_s using a new class called `SettingsMaskedDataset`, which contains the settings that 
customize how a `MaskedDataset` is created. This has two inputs, `data_trim_left` and `data_trim_right`:

- `data_trim_left`:

  The `Dataset`'s `image` and `noise_map` are trimmed and removed from the left (e.g. 1d index values from 0).
  
  For example, if the `Dataset` has shape (100,) and we set `data_trim_left=10`, the `MaskedDataset` that is fitted 
  will have shape (90,). The `mask` is trimmed in the same way.

- `data_trim_right`:

  This behaves the same as `data_trim_left`, however `data` is removed from the right (e.g. 1D index values from the
  shape[0] value of the 1D data).

For our first phase, we will omit both of these settings (by passing them as `None`) and perform the fit from tutorial
4 where we fit a single `Gaussian` profile to `data` composed of a single `Gaussian` (unlike tutorial 4, we'll use a
_CollectionPriorModel_ to do this).
"""

# %%
settings_masked_dataset = htf.SettingsMaskedDataset(
    data_trim_left=None, data_trim_right=None
)

settings = htf.SettingsPhase(settings_masked_dataset=settings_masked_dataset)

phase = htf.Phase(
    phase_name="phase_t7",
    profiles=af.CollectionPriorModel(gaussian=htf.profiles.Gaussian),
    settings=settings,
    search=af.Emcee(),
)

# %%
"""
Import the `simulators` module and set up the `Dataset` and `mask`.
"""

# %%
from howtofit.simulators.chapter_1 import gaussian_x1

dataset = htf.Dataset(data=gaussian_x1.data, noise_map=gaussian_x1.noise_map)
mask = np.full(fill_value=False, shape=dataset.data.shape)

print(
    "Emcee has begun running - checkout the autofit_workspace/howtofit/chapter_1_introduction/output/phase_t6"
    " folder for live output of the results."
    "This Jupyter notebook cell with progress once Emcee has completed - this could take a few minutes!"
)

phase.run(dataset=dataset, mask=mask)

print("Emcee has finished run - you may now continue the notebook.")

# %%
"""
Okay, what happened differently in this _Phase? To begin, lets note the output directory:

`autofit_workspace/howtofit/chapter_1_introduction/tutorial_7_phase_customization/output/phase_t7/settings`

There is a small change to this directory compared to tutorial 6, there is a new folder `settings` within which the
results are stored. It`ll be clear why this is in a moment.

Next, we`re going to customize and run a phase using the `data_trim_left` and `data_trim_right` parameters. First, we 
create a `SettingsMaskedDataset` and `SettingsPhase` object using our input values of these parameters. 
"""

# %%
settings_masked_dataset = htf.SettingsMaskedDataset(
    data_trim_left=20, data_trim_right=30
)

settings = htf.SettingsPhase(settings_masked_dataset=settings_masked_dataset)

# %%
"""
We now create a new `Phase` with these settings and run it (note that we haven`t changed the `phase_name` from 
`phase_t7`, which you might think would cause conflicts in the path the results are output to).
"""

# %%
phase = htf.Phase(
    phase_name="phase_t7",
    profiles=af.CollectionPriorModel(gaussian=htf.profiles.Gaussian),
    settings=settings,
    search=af.Emcee(),
)

print(
    "Emcee has begun running - checkout the autofit_workspace/howtofit/chapter_1_introduction/output/phase_t6"
    " folder for live output of the results."
    "This Jupyter notebook cell with progress once Emcee has completed - this could take a few minutes!"
)

phase.run(dataset=dataset, mask=mask)

print("Emcee has finished run - you may now continue the notebook.")

# %%
"""
You`ll note the results are now in a slightly different directory to the fit performed above:

 `autofit_workspace/howtofit/chapter_1_introduction/tutorial_7_phase_customization/output/phase_example/settings__trim_left_20__trim_right_30`

By customizing the `PhaseSetting``., **PyAutoFit** has changed the output path using a `tag`. There are two reasons 
**PyAutoFit** does this:

 1) Tags describes the analysis, making it explicit what was done to the `Dataset` for the fit.

 2) Tags create a unique output path, allowing you to compare results of `Phase`'s that use different `SettingsPhase`. 
    Equally if you run multiple phases with different `PhaseSetting`'s this ensures the `NonLinearSearch` won`t
    use results generated via a different analysis method.

You should now check out the `settings.py` and `dataset.py` modules, to see how we implemented this.


When reading through this tutorial`s example source code, you may have felt it was a bit clunky having multiple 
_Settings_ classes each of which was set up separately to customize the `MaskedDataset` or the `Phase`. 

For the source code, it actually is quite clunky and could certainly be refactored to make the source code more clean. 
However, through experience we have found this design creates a much better API for a user when choosing settings, 
which will be seen in the next tutorial. Thus, we recommend you adopt the same settings API for your project!


In this tutorial, the `PhaseSetting`'s changed the `MaskedDataset` that was fitted. However, `PhaseSetting`'s do not 
necessarily need to customize the `Dataset`. For example, they could control some aspect of the model, for example the 
precision with which an aspect of the model is numerically calculated. For more complex fitting procedures, they may 
control whether certain features are used, which when turned on / off reduce the accuracy of the model at the expense 
of greater computational run-time.

_PhaseSetting_s are project specific and it could well be your modeling problem is simple enough not to need them.
However, if it does, remember that `PhaseSetting`. are a powerful means to fit models in different ways and compare 
whether this changes the model inferred. In later chapters, we'll discuss more complex model-fitting procedures that 
use `fast` less accurate `PhaseSetting`'s to initialize the model-fit, but switch to slower more accurate 
_PhaseSetting_`s later on.
"""



import numpy as np

"""The `dataset.py` module has been extended to give the `Dataset` a name and metadata."""


class Dataset:
    def __init__(self, data, noise_map, name=None):
        """A class containing the data and noise-map of a 1D line `Dataset`.

        Parameters
        ----------
        data : np.ndarray
            The array of the data, in arbitrary units.
        noise_map : np.ndarray
            An array describing the RMS standard deviation error in each data pixel, in arbitrary units.
        """
        self.data = data
        self.noise_map = noise_map

        # The name of the `Dataset` is used by the aggregator, to determine the name of the file the `Dataset` is saved as
        # and so that when using the aggregator you can know which `Dataset` you are manipulating.

        self.name = name if name is str else "dataset"

    @property
    def xvalues(self):
        return np.arange(self.data.shape[0])

    def with_left_trimmed(self, data_trim_left):

        """Here, we use the existing `MaskedDataset` to create a trimmed `Dataset`."""

        data_trimmed = self.data[data_trim_left:]
        noise_map_trimmed = self.noise_map[data_trim_left:]

        return Dataset(data=data_trimmed, noise_map=noise_map_trimmed)

    def with_right_trimmed(self, data_trim_right):

        """We do the same as above, but removing data to the right."""

        data_trimmed = self.data[:-data_trim_right]
        noise_map_trimmed = self.noise_map[:-data_trim_right]

        return Dataset(data=data_trimmed, noise_map=noise_map_trimmed)


class SettingsMaskedDataset:
    def __init__(self, data_trim_left=None, data_trim_right=None):
        """
        The settings of the `MaskedDataset` class, that in a phase are used to deterimne if the `MaskedDataset` is
        trimmed from the left and / or right before model-fitting.

        This class includes tags which are used to customize the output folders of a run dependent on the settings.

        Parameters
        ----------
        data_trim_left : int or None
            The number of pixels in 1D from the left (NumPy index 0) that the `Dataset` is trimmed.
        data_trim_right : int or None
            The number of pixels in 1D from the right (NumPy index -1) that the `Dataset` is trimmed.
        """

        self.data_trim_left = data_trim_left
        self.data_trim_right = data_trim_right

    @property
    def tag(self):
        """Generate a tag describin all settings customizing the `MaskedDataset`, which for this example only describes
        how the dataset it trimmed from the left and right.
        """
        return f"{self.data_trim_left_tag}{self.data_trim_right_tag}"

    @property
    def data_trim_left_tag(self):
        """Generate a data trim left tag, to customize phase names based on how much of the `Dataset` is trimmed to
        its left.

        This changes the phase name `settings` as follows:

        data_trim_left = None -> settings
        data_trim_left = 2 -> settings__trim_left_2
        data_trim_left = 10 -> settings__trim_left_10
        """
        if self.data_trim_left is None:
            return ""
        return f"__trim_left_{str(self.data_trim_left)}"

    @property
    def data_trim_right_tag(self):
        """Generate a data trim right tag, to customize phase names based on how much of the `Dataset` is trimmed to its right.

        This changes the phase name `settings` as follows:

        data_trim_right = None -> settings
        data_trim_right = 2 -> settings__trim_right_2
        data_trim_right = 10 -> settings__trim_right_10
        """
        if self.data_trim_right is None:
            return ""
        return f"__trim_right_{str(self.data_trim_right)}"


class MaskedDataset:
    def __init__(self, dataset, mask, settings=SettingsMaskedDataset()):
        """
        A masked dataset, which is an image, noise-map and mask.

        Parameters
        ----------
        dataset: im.Dataset
            The `Dataset` (the image, noise-map, etc.)
        mask: msk.Mask2D
            The 1D mask that is applied to the `Dataset`.
        """

        if settings.data_trim_left is not None:
            dataset = dataset.with_left_trimmed(data_trim_left=settings.data_trim_left)
            mask = mask[settings.data_trim_left :]

        if settings.data_trim_right is not None:
            dataset = dataset.with_right_trimmed(
                data_trim_right=settings.data_trim_right
            )
            mask = mask[: -settings.data_trim_right]

        self.dataset = dataset
        self.mask = mask
        self.data = dataset.data * np.invert(mask)
        self.noise_map = dataset.noise_map * np.invert(mask)

    @property
    def xvalues(self):
        return np.arange(self.data.shape[0])

    def signal_to_noise_map(self):
        return self.data / self.noise_map
